"""PDF / DOCX / XLSX / PPTX 与图片提取为 TXT、HTML、Markdown；图片外置到 images_<源文件名> 目录。"""

from __future__ import annotations

import html
import io
import re
import shutil
from pathlib import Path
from typing import Any, Callable, Iterator

_VALID_FORMATS = frozenset({"markdown", "html", "txt"})


def _images_subdir_name(stem: str) -> str:
    return f"images_{stem}"


def _ensure_images_dir(output_dir: Path, stem: str) -> Path:
    d = output_dir / _images_subdir_name(stem)
    d.mkdir(parents=True, exist_ok=True)
    return d


def _page_index_range(
    total: int, page_start: int | None, page_end: int | None
) -> range:
    """1-based 闭区间 → 0-based range；total 为页/幻灯片总数。"""
    if total <= 0:
        return range(0, 0)
    a = 1 if page_start is None else max(1, int(page_start))
    b = total if page_end is None else int(page_end)
    b = min(total, max(a, b))
    return range(a - 1, b)


def _ocr_pdf_page(page: Any) -> tuple[str | None, str | None]:
    """返回 (文本, 错误说明)。需安装可选依赖 pytesseract 与系统 Tesseract。"""
    try:
        import pytesseract
        from PIL import Image
    except ImportError as e:
        return None, f"OCR skipped: {e} (pip install pytesseract, install Tesseract-OCR binary)"
    import fitz

    m = fitz.Matrix(2.0, 2.0)
    pix = page.get_pixmap(matrix=m, alpha=False)
    img = Image.open(io.BytesIO(pix.tobytes("png")))
    try:
        text = pytesseract.image_to_string(img, lang="eng+chi_sim")
    except Exception as e:  # noqa: BLE001
        el = repr(e).lower()
        if "tesseract" in el and ("not" in el or "path" in el or "cmd" in el):
            return None, "Tesseract executable not found (install Tesseract-OCR and PATH)"
        return None, repr(e)
    return (text or "").strip() or None, None


def _markdown_to_txt(md: str) -> str:
    t = re.sub(r"^#+\s*", "", md, flags=re.MULTILINE)
    t = re.sub(r"!\[[^\]]*\]\(([^)]+)\)", r"[Image: \1]", t)
    return t.strip()


def _markdown_subset_to_html(md: str, title: str) -> str:
    parts: list[str] = [
        "<!DOCTYPE html>",
        '<html lang="zh-CN">',
        "<head>",
        '<meta charset="utf-8"/>',
        f"<title>{html.escape(title)}</title>",
        "</head><body>",
    ]
    for block in re.split(r"\n\n+", md.strip()):
        block = block.strip()
        if not block:
            continue
        if block.startswith("## "):
            parts.append(f"<h2>{html.escape(block[3:].strip())}</h2>")
        elif block.startswith("![](") and block.endswith(")"):
            src = block[4:-1].strip()
            parts.append(
                f'<p><img src="{html.escape(src, quote=True)}" alt=""/></p>'
            )
        elif block.startswith("# "):
            parts.append(f"<h1>{html.escape(block[2:].strip())}</h1>")
        else:
            for ln in block.split("\n"):
                parts.append(f"<p>{html.escape(ln)}</p>")
    parts.append("</body></html>")
    return "\n".join(parts)


class DocumentConverter:
    def __init__(self, policy_check: Callable[[Path], None]) -> None:
        self._check = policy_check

    def convert_to_markdown(self, source: str, output_dir: str) -> dict[str, Any]:
        """兼容旧工具：等价于 doc_extract format=markdown，全页/全表。保留 markdown / assets 字段名。"""
        r = self.convert_document(
            source,
            output_dir,
            output_format="markdown",
            page_start=None,
            page_end=None,
            use_ocr=False,
            sheet_names=None,
        )
        if r.get("ok"):
            r["markdown"] = r["output_path"]
            r["assets"] = r["images_dir"]
        return r

    def convert_document(
        self,
        source: str,
        output_dir: str,
        *,
        output_format: str = "markdown",
        page_start: int | None = None,
        page_end: int | None = None,
        use_ocr: bool = False,
        sheet_names: list[str] | None = None,
    ) -> dict[str, Any]:
        fmt = (output_format or "markdown").strip().lower()
        if fmt not in _VALID_FORMATS:
            return {"ok": False, "error": f"output_format must be one of {sorted(_VALID_FORMATS)}"}

        src = Path(source).expanduser().resolve()
        out = Path(output_dir).expanduser().resolve()
        self._check(src)
        self._check(out)
        out.mkdir(parents=True, exist_ok=True)

        stem = src.stem
        images_dir = _ensure_images_dir(out, stem)
        rel_img_prefix = images_dir.name

        suf = src.suffix.lower()
        notes: list[str] = []

        if page_start is not None or page_end is not None:
            if suf not in {".pdf", ".pptx"}:
                notes.append(
                    "page_start/page_end apply to PDF and PPTX only; ignored for this type."
                )

        if suf == ".pdf":
            body = self._extract_pdf(
                src,
                images_dir,
                rel_img_prefix,
                page_start,
                page_end,
                use_ocr,
                notes,
            )
        elif suf == ".docx" and fmt == "html":
            final = self._extract_docx_html(src, images_dir, rel_img_prefix, notes)
            out_path = out / f"{stem}.html"
            out_path.write_text(final, encoding="utf-8")
            return {
                "ok": True,
                "output_path": str(out_path),
                "output_format": "html",
                "images_dir": str(images_dir),
                "images_relative_prefix": rel_img_prefix,
                "notes": notes,
            }
        elif suf == ".docx":
            try:
                body = self._extract_docx_markdown(
                    src, images_dir, rel_img_prefix, notes
                )
            except Exception as e:  # noqa: BLE001
                notes.append(
                    f"mammoth DOCX conversion failed ({type(e).__name__}: {e}); "
                    "falling back to python-docx plain text (tables as tab-separated rows; embedded images not inlined)."
                )
                body = self._extract_docx_plain_python_docx(src, notes)
        elif suf == ".xlsx":
            body = self._extract_xlsx(
                src, page_start, page_end, sheet_names, notes
            )
        elif suf == ".pptx":
            body = self._extract_pptx(
                src,
                images_dir,
                rel_img_prefix,
                page_start,
                page_end,
                notes,
            )
        elif suf in {".png", ".jpg", ".jpeg", ".gif", ".webp"}:
            body = self._extract_image_file(src, images_dir, rel_img_prefix)
        else:
            return {"ok": False, "error": f"unsupported type: {suf}"}

        if not body.startswith("#"):
            body = f"# {src.name}\n\n" + body

        ext = {"markdown": ".md", "html": ".html", "txt": ".txt"}[fmt]
        out_path = out / f"{stem}{ext}"

        if fmt == "markdown":
            final = body
        elif fmt == "txt":
            final = _markdown_to_txt(body)
        else:
            final = _markdown_subset_to_html(body, stem)

        out_path.write_text(final, encoding="utf-8")

        return {
            "ok": True,
            "output_path": str(out_path),
            "output_format": fmt,
            "images_dir": str(images_dir),
            "images_relative_prefix": rel_img_prefix,
            "notes": notes,
        }

    def _extract_pdf(
        self,
        src: Path,
        images_dir: Path,
        rel_prefix: str,
        page_start: int | None,
        page_end: int | None,
        use_ocr: bool,
        notes: list[str],
    ) -> str:
        import fitz

        doc = fitz.open(src)
        try:
            pr = _page_index_range(len(doc), page_start, page_end)
            parts: list[str] = [f"# {src.name}\n"]
            img_count = 0
            for page_index in pr:
                page = doc[page_index]
                pn = page_index + 1
                parts.append(f"## Page {pn}\n")
                text = (page.get_text("text") or "").strip()
                if use_ocr:
                    ocr_t, ocr_err = _ocr_pdf_page(page)
                    if ocr_err:
                        notes.append(f"page {pn}: {ocr_err}")
                    if ocr_t:
                        if text:
                            parts.append(text + "\n\n")
                            parts.append("### OCR\n" + ocr_t + "\n")
                        else:
                            parts.append(ocr_t + "\n")
                    elif not text:
                        parts.append("_(no text layer; OCR unavailable or empty)_\n")
                else:
                    parts.append((text or "_(no text)_") + "\n")
                for img in page.get_images(full=True):
                    xref = img[0]
                    base = doc.extract_image(xref)
                    img_count += 1
                    name = f"pdf_p{pn}_{img_count}.{base['ext']}"
                    dest = images_dir / name
                    dest.write_bytes(base["image"])
                    parts.append(f"![]({rel_prefix}/{name})\n")
            return "\n".join(parts)
        finally:
            doc.close()

    def _make_docx_image_converter(
        self, images_dir: Path, rel_prefix: str
    ) -> Callable[..., dict[str, str]]:
        counter = [0]

        def convert_image(image: Any) -> dict[str, str]:
            with image.open() as f:
                data = f.read()
            counter[0] += 1
            ct = (image.content_type or "image/png").lower()
            if "jpeg" in ct or "jpg" in ct:
                ext = "jpg"
            elif "png" in ct:
                ext = "png"
            elif "gif" in ct:
                ext = "gif"
            else:
                ext = "bin"
            name = f"docx_{counter[0]}.{ext}"
            dest = images_dir / name
            dest.write_bytes(data)
            return {"src": f"{rel_prefix}/{name}"}

        return convert_image

    def _extract_docx_plain_python_docx(self, src: Path, notes: list[str]) -> str:
        """
        Mammoth 失败或异常文档时的回退：按文档顺序抽取段落与表格为纯文本（表格行为制表符分隔）。
        不导出内嵌图片路径；复杂排版可能弱于 Mammoth。
        """
        from docx import Document
        from docx.oxml.table import CT_Tbl
        from docx.oxml.text.paragraph import CT_P
        from docx.table import Table, _Cell
        from docx.text.paragraph import Paragraph

        doc = Document(str(src))
        lines: list[str] = []

        def iter_block_items(parent: Any) -> Iterator[Any]:
            if isinstance(parent, Document):
                parent_elm = parent.element.body
            elif isinstance(parent, _Cell):
                parent_elm = parent._tc
            else:
                return
            for child in parent_elm.iterchildren():
                if isinstance(child, CT_P):
                    yield Paragraph(child, parent)
                elif isinstance(child, CT_Tbl):
                    yield Table(child, parent)

        try:
            for block in iter_block_items(doc):
                if isinstance(block, Paragraph):
                    t = (block.text or "").strip()
                    if t:
                        lines.append(t)
                elif isinstance(block, Table):
                    for row in block.rows:
                        cells = [(c.text or "").strip() for c in row.cells]
                        if any(cells):
                            lines.append("\t".join(cells))
        except Exception as e:  # noqa: BLE001
            notes.append(f"python-docx plain fallback partial error: {e!r}")
            for para in doc.paragraphs:
                t = (para.text or "").strip()
                if t:
                    lines.append(t)
            for table in doc.tables:
                for row in table.rows:
                    cells = [(c.text or "").strip() for c in row.cells]
                    if any(cells):
                        lines.append("\t".join(cells))

        text = "\n\n".join(lines).strip()
        return text if text else "_(no extractable text)_"

    def _extract_docx_markdown(
        self,
        src: Path,
        images_dir: Path,
        rel_prefix: str,
        notes: list[str],
    ) -> str:
        import mammoth

        cv = self._make_docx_image_converter(images_dir, rel_prefix)
        with src.open("rb") as f:
            result = mammoth.convert_to_markdown(f, convert_image=cv)
        notes.extend(m.message for m in result.messages)
        return result.value

    def _extract_docx_html(
        self,
        src: Path,
        images_dir: Path,
        rel_prefix: str,
        notes: list[str],
    ) -> str:
        import mammoth

        cv = self._make_docx_image_converter(images_dir, rel_prefix)
        with src.open("rb") as f:
            result = mammoth.convert_to_html(f, convert_image=cv)
        notes.extend(m.message for m in result.messages)
        return result.value

    def _extract_xlsx(
        self,
        src: Path,
        page_start: int | None,
        page_end: int | None,
        sheet_names: list[str] | None,
        notes: list[str],
    ) -> str:
        import openpyxl

        wb = openpyxl.load_workbook(src, read_only=True, data_only=True)
        try:
            names = list(wb.sheetnames)
            if sheet_names:
                sheets = [n for n in names if n in set(sheet_names)]
                if not sheets:
                    notes.append("sheet_names matched nothing; using all sheets.")
                    sheets = names
            else:
                pr = _page_index_range(len(names), page_start, page_end)
                sheets = [names[i] for i in pr]
            parts: list[str] = [f"# {src.stem}\n"]
            for sheet in sheets:
                ws = wb[sheet]
                parts.append(f"\n## {sheet}\n")
                rows_out: list[str] = []
                for row in ws.iter_rows(values_only=True):
                    cells = ["" if c is None else str(c) for c in row]
                    if any(x.strip() for x in cells):
                        rows_out.append("| " + " | ".join(cells) + " |")
                if rows_out:
                    parts.append("\n".join(rows_out))
            return "\n".join(parts)
        finally:
            wb.close()

    def _extract_pptx(
        self,
        src: Path,
        images_dir: Path,
        rel_prefix: str,
        page_start: int | None,
        page_end: int | None,
        notes: list[str],
    ) -> str:
        from pptx import Presentation
        from pptx.enum.shapes import MSO_SHAPE_TYPE

        prs = Presentation(str(src))
        nslides = len(prs.slides)
        pr = _page_index_range(nslides, page_start, page_end)
        parts: list[str] = [f"# {src.stem}\n"]
        pic_n = 0
        for si in pr:
            slide = prs.slides[si]
            sn = si + 1
            parts.append(f"\n## Slide {sn}\n")
            for shape in slide.shapes:
                if getattr(shape, "has_text_frame", False):
                    for para in shape.text_frame.paragraphs:
                        t = para.text.strip()
                        if t:
                            parts.append(t + "\n")
                if shape.shape_type == MSO_SHAPE_TYPE.PICTURE:
                    try:
                        blob = shape.image.blob
                        ext = (shape.image.ext or "png").lstrip(".")
                        pic_n += 1
                        name = f"pptx_s{sn}_{pic_n}.{ext}"
                        dest = images_dir / name
                        dest.write_bytes(blob)
                        parts.append(f"![]({rel_prefix}/{name})\n")
                    except Exception as e:  # noqa: BLE001
                        notes.append(f"slide {sn} picture: {e!r}")
        return "\n\n".join(parts)

    def _extract_image_file(
        self, src: Path, images_dir: Path, rel_prefix: str
    ) -> str:
        from PIL import Image

        dest = images_dir / src.name
        shutil.copy2(src, dest)
        with Image.open(dest) as im:
            info = {"format": im.format, "size": im.size, "mode": im.mode}
        return (
            f"# Image {src.name}\n\n"
            f"![]({rel_prefix}/{dest.name})\n\n"
            f"Metadata: `{info}`\n"
        )



"""chat_serialize：多模态消息 token 粗估。"""

from __future__ import annotations

from aicd.agent.chat_serialize import (
    VISION_IMAGE_TOKENS_ESTIMATE,
    estimate_message_tokens,
)


def test_estimate_message_tokens_plain_string() -> None:
    msg = {"role": "user", "content": "abcd"}
    assert estimate_message_tokens(msg) == 2  # 4 // 4 + 1


def test_estimate_message_tokens_multimodal_one_image() -> None:
    msg = {
        "role": "user",
        "content": [
            {"type": "text", "text": "x"},
            {"type": "image_url", "image_url": {"url": "data:image/png;base64,..."}},
        ],
    }
    # text: 1 char -> 1//4+1 = 1; one image -> + VISION_IMAGE_TOKENS_ESTIMATE
    assert estimate_message_tokens(msg) == 1 + VISION_IMAGE_TOKENS_ESTIMATE


def test_estimate_message_tokens_multimodal_two_images() -> None:
    msg = {
        "role": "user",
        "content": [
            {"type": "image_url", "image_url": {"url": "http://a"}},
            {"type": "image_url", "image_url": {"url": "http://b"}},
        ],
    }
    assert estimate_message_tokens(msg) == 2 * VISION_IMAGE_TOKENS_ESTIMATE



"""agent / llm 采样字段 YAML 往返。"""

from __future__ import annotations

from pathlib import Path

from aicd.config import AppConfig, load_config, save_config
from aicd.config_settings import apply_nested_settings_patch, format_settings_summary


def test_yaml_roundtrip_agent_llm(tmp_path: Path) -> None:
    p = tmp_path / "c.yaml"
    cfg = AppConfig()
    cfg.agent.max_tool_rounds = 40
    cfg.agent.sub_agent_max_tool_rounds = 8
    cfg.llm.temperature = 0.2
    cfg.llm.max_output_tokens = 2048
    cfg.llm.frequency_penalty = 0.1
    cfg.llm.extra_body = {"repeat_penalty": 1.05}
    save_config(p, cfg)
    cfg2 = load_config(p)
    assert cfg2.agent.max_tool_rounds == 40
    assert cfg2.agent.sub_agent_max_tool_rounds == 8
    assert cfg2.llm.temperature == 0.2
    assert cfg2.llm.max_output_tokens == 2048
    assert cfg2.llm.frequency_penalty == 0.1
    assert cfg2.llm.extra_body.get("repeat_penalty") == 1.05


def test_apply_patch_tool() -> None:
    cfg = AppConfig()
    ok, msg = apply_nested_settings_patch(
        cfg,
        {"llm": {"temperature": 0.5}, "agent": {"maxToolRounds": 10}},
    )
    assert ok and msg == "ok"
    assert cfg.llm.temperature == 0.5
    assert cfg.agent.max_tool_rounds == 10
    assert "max_tool_rounds=10" in format_settings_summary(cfg)


def test_apply_patch_invalid_restores() -> None:
    cfg = AppConfig()
    cfg.llm.temperature = 0.1
    ok, _ = apply_nested_settings_patch(cfg, {"llm": {"temperature": 99}})
    assert not ok
    assert cfg.llm.temperature == 0.1


def test_tool_rounds_clamped_to_max(tmp_path: Path) -> None:
    p = tmp_path / "c.yaml"
    p.write_text(
        "agent:\n  maxToolRounds: 800\n  subAgentMaxToolRounds: 1200\n",
        encoding="utf-8",
    )
    cfg = load_config(p)
    assert cfg.agent.max_tool_rounds == 500
    assert cfg.agent.sub_agent_max_tool_rounds == 500


def test_apply_patch_tool_rounds_clamped() -> None:
    cfg = AppConfig()
    ok, msg = apply_nested_settings_patch(cfg, {"agent": {"maxToolRounds": 500}})
    assert ok and msg == "ok"
    assert cfg.agent.max_tool_rounds == 500
    ok2, _ = apply_nested_settings_patch(cfg, {"agent": {"maxToolRounds": 600}})
    assert ok2
    assert cfg.agent.max_tool_rounds == 500


def test_yaml_llm_retry_roundtrip(tmp_path: Path) -> None:
    p = tmp_path / "c.yaml"
    p.write_text(
        "llm:\n  maxRetries: 5\n  retryBaseDelaySec: 2.5\n",
        encoding="utf-8",
    )
    cfg = load_config(p)
    assert cfg.llm.max_retries == 5
    assert cfg.llm.retry_base_delay_sec == 2.5



"""使用 openpyxl / python-pptx 写入 xlsx、pptx。"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

from openpyxl import Workbook, load_workbook
from pptx import Presentation
from pptx.chart.data import CategoryChartData
from pptx.dml.color import RGBColor
from pptx.enum.chart import XL_CHART_TYPE
from pptx.enum.shapes import MSO_SHAPE
from pptx.oxml import parse_xml
from pptx.oxml.ns import qn
from pptx.util import Inches, Pt

MAX_SHEETS = 20
MAX_ROWS_PER_SHEET = 5000
MAX_COLS = 256
MAX_SLIDES = 40
MAX_BULLETS = 24
MAX_PPTX_DECOR_SHAPES = 16
MAX_PPTX_TABLE_ROWS = 24
MAX_PPTX_TABLE_COLS = 16
MAX_PPTX_CHART_SERIES = 6
MAX_PPTX_CHART_CATEGORIES = 32
MAX_PPTX_TEXT_BOXES = 12
MAX_PPTX_IMAGES_PER_SLIDE = 12
MAX_PPTX_SPEAKER_NOTES_CHARS = 16000
MAX_PPTX_TEXT_BOX_LINES = 32
_RASTER_IMG_EXT = {".png", ".jpg", ".jpeg", ".gif", ".bmp", ".tif", ".tiff"}

_PPTX_SHAPE_ALIASES: dict[str, MSO_SHAPE] = {
    "rectangle": MSO_SHAPE.RECTANGLE,
    "rounded_rectangle": MSO_SHAPE.ROUNDED_RECTANGLE,
    "oval": MSO_SHAPE.OVAL,
}

# Returned by pptx_inspect_template so callers can pick richer auto shapes without reading python-pptx docs.
_PPTX_SHAPE_KIND_EXAMPLES: list[str] = [
    "chevron",
    "diamond",
    "pentagon",
    "hexagon",
    "star_5_point",
    "right_arrow",
    "down_arrow",
    "left_brace",
    "cloud",
    "heart",
    "lightning_bolt",
    "sun",
    "moon",
    "arc",
    "can",
    "cube",
    "folded_corner",
    "plaque",
    "curved_up_ribbon",
    "round_1_rectangle",
]

_PPTX_CHART_TYPES: dict[str, XL_CHART_TYPE] = {
    "column_clustered": XL_CHART_TYPE.COLUMN_CLUSTERED,
    "column_stacked": XL_CHART_TYPE.COLUMN_STACKED,
    "column_stacked_100": XL_CHART_TYPE.COLUMN_STACKED_100,
    "line": XL_CHART_TYPE.LINE_MARKERS,
    "line_stacked": XL_CHART_TYPE.LINE_STACKED,
    "line_stacked_100": XL_CHART_TYPE.LINE_STACKED_100,
    "pie": XL_CHART_TYPE.PIE,
    "bar_clustered": XL_CHART_TYPE.BAR_CLUSTERED,
    "bar_stacked": XL_CHART_TYPE.BAR_STACKED,
    "area": XL_CHART_TYPE.AREA,
    "area_stacked": XL_CHART_TYPE.AREA_STACKED,
    "doughnut": XL_CHART_TYPE.DOUGHNUT,
    "radar": XL_CHART_TYPE.RADAR,
}


def _pptx_mso_shape_from_kind(kind: str) -> MSO_SHAPE | None:
    k = kind.strip()
    if not k:
        return None
    low = k.lower()
    if low in _PPTX_SHAPE_ALIASES:
        return _PPTX_SHAPE_ALIASES[low]
    enum_name = k.upper().replace("-", "_")
    return getattr(MSO_SHAPE, enum_name, None)


def _pptx_slide_add_shapes(slide: Any, shapes_spec: Any) -> str | None:
    """Optional accent shapes; new shapes stack above layout placeholders—keep positions in margins."""
    if shapes_spec is None:
        return None
    if not isinstance(shapes_spec, list):
        return "shapes must be a list"
    if len(shapes_spec) > MAX_PPTX_DECOR_SHAPES:
        return f"max {MAX_PPTX_DECOR_SHAPES} shapes per slide"
    for i, s in enumerate(shapes_spec):
        if not isinstance(s, dict):
            return f"shapes[{i}] must be object"
        kind = str(s.get("kind", "rectangle")).strip()
        mso = _pptx_mso_shape_from_kind(kind)
        if mso is None:
            return (
                f"shapes[{i}].kind {kind!r} unknown — use lowercase aliases "
                f"{sorted(_PPTX_SHAPE_ALIASES)} or any **MSO_SHAPE** enum name "
                f"(e.g. {_PPTX_SHAPE_KIND_EXAMPLES[:8]!r}, …); see **pptx_inspect_template** `shape_kind_examples`."
            )
        try:
            left = Inches(float(s["left_in"]))
            top = Inches(float(s["top_in"]))
            w = Inches(float(s["width_in"]))
            h = Inches(float(s["height_in"]))
        except (KeyError, TypeError, ValueError):
            return f"shapes[{i}] needs numeric left_in, top_in, width_in, height_in"
        shape = slide.shapes.add_shape(mso, left, top, w, h)
        fill = s.get("fill_rgb")
        if fill is not None:
            if (
                not isinstance(fill, (list, tuple))
                or len(fill) != 3
                or not all(isinstance(x, (int, float)) for x in fill)
            ):
                return f"shapes[{i}].fill_rgb must be [R,G,B] integers 0–255"
            r, g, b = (max(0, min(255, int(x))) for x in fill)
            shape.fill.solid()
            shape.fill.fore_color.rgb = RGBColor(r, g, b)
        line = s.get("line_rgb")
        if line is not None:
            if (
                not isinstance(line, (list, tuple))
                or len(line) != 3
                or not all(isinstance(x, (int, float)) for x in line)
            ):
                return f"shapes[{i}].line_rgb must be [R,G,B] integers 0–255"
            r, g, b = (max(0, min(255, int(x))) for x in line)
            shape.line.color.rgb = RGBColor(r, g, b)
    return None


def _pptx_slide_add_table(slide: Any, table_spec: Any) -> str | None:
    if table_spec is None:
        return None
    if not isinstance(table_spec, dict):
        return "table must be object"
    data = table_spec.get("data")
    if not isinstance(data, list) or not data:
        return "table.data must be non-empty list of rows"
    if len(data) > MAX_PPTX_TABLE_ROWS:
        return f"table max {MAX_PPTX_TABLE_ROWS} rows"
    norm: list[list[str]] = []
    cols = 0
    for ri, row in enumerate(data):
        if isinstance(row, (list, tuple)):
            cells = [str(c) if c is not None else "" for c in row]
        else:
            cells = [str(row)]
        cols = max(cols, len(cells))
        norm.append(cells)
    if cols > MAX_PPTX_TABLE_COLS:
        return f"table max {MAX_PPTX_TABLE_COLS} columns"
    for ri, row in enumerate(norm):
        while len(row) < cols:
            row.append("")
        norm[ri] = row[:cols]
    try:
        left = Inches(float(table_spec.get("table_left_in", 0.5)))
        top = Inches(float(table_spec.get("table_top_in", 4.0)))
        tw = float(table_spec.get("table_width_in", 9.0))
        th = float(table_spec.get("table_height_in", 2.0))
    except (TypeError, ValueError):
        return "table needs numeric table_left_in, table_top_in, table_width_in, table_height_in"
    rows_n = len(norm)
    gf = slide.shapes.add_table(rows_n, cols, left, top, Inches(tw), Inches(th))
    tbl = gf.table
    for ri in range(rows_n):
        for ci in range(cols):
            tbl.cell(ri, ci).text = norm[ri][ci]
    return None


def _pptx_slide_add_chart(slide: Any, chart_spec: Any) -> str | None:
    if chart_spec is None:
        return None
    if not isinstance(chart_spec, dict):
        return "chart must be object"
    ct_key = str(chart_spec.get("chart_type", "column_clustered")).lower().replace("-", "_")
    xl_type = _PPTX_CHART_TYPES.get(ct_key)
    if xl_type is None:
        return f"chart.chart_type {ct_key!r} unknown; use {sorted(_PPTX_CHART_TYPES)}"
    categories = chart_spec.get("categories") or []
    series_list = chart_spec.get("series") or []
    if not isinstance(categories, list) or not categories:
        return "chart.categories must be non-empty list"
    if len(categories) > MAX_PPTX_CHART_CATEGORIES:
        return f"chart max {MAX_PPTX_CHART_CATEGORIES} categories"
    if not isinstance(series_list, list) or not series_list:
        return "chart.series must be non-empty list of {name, values}"
    if ct_key in ("pie", "doughnut") and len(series_list) != 1:
        return "pie and doughnut charts require exactly one entry in chart.series"
    if len(series_list) > MAX_PPTX_CHART_SERIES:
        return f"chart max {MAX_PPTX_CHART_SERIES} series"
    n = len(categories)
    chart_data = CategoryChartData()
    chart_data.categories = [str(c) for c in categories]
    for j, ser in enumerate(series_list):
        if not isinstance(ser, dict):
            return f"chart.series[{j}] must be object"
        name = str(ser.get("name") or f"Series {j + 1}")
        vals = ser.get("values")
        if not isinstance(vals, (list, tuple)):
            return f"chart.series[{j}].values must be list of numbers"
        if len(vals) != n:
            return f"chart.series[{j!r}].values length must match categories ({n})"
        chart_data.add_series(name, tuple(vals))
    try:
        left = Inches(float(chart_spec.get("chart_left_in", 0.5)))
        top = Inches(float(chart_spec.get("chart_top_in", 1.5)))
        cw = float(chart_spec.get("chart_width_in", 9.0))
        ch = float(chart_spec.get("chart_height_in", 4.5))
    except (TypeError, ValueError):
        return "chart needs numeric chart_left_in, chart_top_in, chart_width_in, chart_height_in"
    gf = slide.shapes.add_chart(xl_type, left, top, Inches(cw), Inches(ch), chart_data)
    chart = gf.chart
    ctitle = chart_spec.get("chart_title")
    if ctitle is not None and str(ctitle).strip():
        chart.has_title = True
        chart.chart_title.text_frame.text = str(ctitle)
    return None


def _pptx_slide_add_text_boxes(slide: Any, spec_list: Any) -> str | None:
    if spec_list is None:
        return None
    if not isinstance(spec_list, list):
        return "text_boxes must be a list"
    if len(spec_list) > MAX_PPTX_TEXT_BOXES:
        return f"max {MAX_PPTX_TEXT_BOXES} text_boxes per slide"
    for i, s in enumerate(spec_list):
        if not isinstance(s, dict):
            return f"text_boxes[{i}] must be object"
        try:
            left = Inches(float(s["left_in"]))
            top = Inches(float(s["top_in"]))
            w = Inches(float(s["width_in"]))
            h = Inches(float(s["height_in"]))
        except (KeyError, TypeError, ValueError):
            return f"text_boxes[{i}] needs left_in, top_in, width_in, height_in (numbers)"
        text = s.get("text")
        lines = s.get("lines")
        if lines is not None:
            if not isinstance(lines, list):
                return f"text_boxes[{i}].lines must be list of strings"
            lines = [str(x) for x in lines[:MAX_PPTX_TEXT_BOX_LINES]]
            body = "\n".join(lines) if lines else ""
        elif text is not None:
            body = str(text)
        else:
            return f"text_boxes[{i}] needs text or lines[]"
        box = slide.shapes.add_textbox(left, top, w, h)
        tf = box.text_frame
        tf.clear()
        tf.word_wrap = True
        parts = body.split("\n")
        if parts:
            tf.text = parts[0]
            for line in parts[1:]:
                p = tf.add_paragraph()
                p.text = line
        fp = s.get("font_pt")
        if fp is not None:
            try:
                pt = float(fp)
            except (TypeError, ValueError):
                return f"text_boxes[{i}].font_pt must be a number"
            for para in tf.paragraphs:
                para.font.size = Pt(pt)
        if s.get("bold") is True:
            for para in tf.paragraphs:
                para.font.bold = True
    return None


def _pptx_picture_err(ip: Path, suf: str) -> dict[str, Any] | None:
    if suf in {".html", ".htm"}:
        return {
            "ok": False,
            "error": "image_path cannot be .html — PowerPoint needs a raster file (.png/.jpg/…).",
            "hint": "Call **viz_html_to_png** with the same html_path (the .html and res/ stay on disk), "
            "then set image_path to the returned png_path.",
        }
    if suf not in _RASTER_IMG_EXT:
        return {
            "ok": False,
            "error": f"unsupported image extension {suf!r}; use {sorted(_RASTER_IMG_EXT)}",
        }
    return None


def _pptx_slide_add_pictures(
    slide: Any,
    prs: Any,
    pics: list[dict[str, Any]],
    policy_check: Callable[[Path], None],
) -> dict[str, Any] | str | None:
    """Return error dict for fatal tool shape, or str for simple error, or None."""
    if not pics:
        return None
    if len(pics) > MAX_PPTX_IMAGES_PER_SLIDE:
        return f"max {MAX_PPTX_IMAGES_PER_SLIDE} images per slide (images[] + image_path)"
    slide_w_in = prs.slide_width / 914400
    for i, pic in enumerate(pics):
        if not isinstance(pic, dict):
            return f"images[{i}] must be object"
        raw = pic.get("image_path")
        if not raw:
            return f"images[{i}].image_path required"
        ip = Path(str(raw)).expanduser().resolve()
        policy_check(ip)
        if not ip.is_file():
            return f"image not found: {ip}"
        suf = ip.suffix.lower()
        pe = _pptx_picture_err(ip, suf)
        if pe is not None:
            return pe
        try:
            left = float(pic.get("image_left_in", 0.5))
            top = float(pic.get("image_top_in", 4.0))
            width = pic.get("image_width_in")
            if width is None:
                width = min(4.0, slide_w_in - left - 0.5)
            else:
                width = float(width)
            height = pic.get("image_height_in")
        except (TypeError, ValueError):
            return f"images[{i}] needs numeric image_left_in, image_top_in, image_width_in"
        kw: dict[str, Any] = {"width": Inches(width)}
        if height is not None:
            kw["height"] = Inches(float(height))
        slide.shapes.add_picture(str(ip), Inches(left), Inches(top), **kw)
    return None


def _pptx_collect_slide_picture_specs(spec: dict[str, Any]) -> list[dict[str, Any]]:
    out: list[dict[str, Any]] = []
    for im in spec.get("images") or []:
        if isinstance(im, dict) and im.get("image_path"):
            out.append(dict(im))
    if spec.get("image_path"):
        out.append(
            {
                "image_path": spec["image_path"],
                "image_left_in": spec.get("image_left_in", 0.5),
                "image_top_in": spec.get("image_top_in", 4.0),
                "image_width_in": spec.get("image_width_in"),
                "image_height_in": spec.get("image_height_in"),
            }
        )
    return out


def _pptx_slide_set_speaker_notes(slide: Any, notes_spec: Any) -> str | None:
    if notes_spec is None:
        return None
    text = str(notes_spec).strip() if not isinstance(notes_spec, str) else notes_spec.strip()
    if not text:
        return None
    if len(text) > MAX_PPTX_SPEAKER_NOTES_CHARS:
        return f"speaker_notes max {MAX_PPTX_SPEAKER_NOTES_CHARS} characters"
    ns = slide.notes_slide
    ntf = ns.notes_text_frame
    if ntf is None:
        return "notes master has no notes body placeholder; cannot set speaker_notes"
    ntf.text = text
    return None


def _pptx_slide_set_transition(slide: Any, tr_spec: Any) -> str | None:
    """Slide transition (between-slide), not per-shape animation."""
    if tr_spec is None:
        return None
    if not isinstance(tr_spec, dict):
        return "slide_transition must be object"
    effect = str(tr_spec.get("effect", "fade")).lower().replace("-", "_")
    speed = str(tr_spec.get("speed", "med")).lower()
    if speed not in {"slow", "med", "fast"}:
        return "slide_transition.speed must be slow|med|fast"
    direction = str(tr_spec.get("direction", "l")).lower()
    if direction not in {"l", "r", "u", "d"}:
        return "slide_transition.direction must be l|r|u|d"
    orient = str(tr_spec.get("orient", "horz")).lower()
    if orient not in {"horz", "vert"}:
        return "slide_transition.orient must be horz|vert (for split)"

    if effect == "fade":
        inner = "<p:fade/>"
    elif effect == "cut":
        inner = "<p:cut/>"
    elif effect == "push":
        inner = f'<p:push dir="{direction}"/>'
    elif effect == "wipe":
        inner = f'<p:wipe dir="{direction}"/>'
    elif effect == "split":
        inner = f'<p:split orient="{orient}"/>'
    elif effect in ("random_bar", "randombar"):
        inner = "<p:randomBar/>"
    elif effect == "cover":
        inner = f'<p:cover dir="{direction}"/>'
    elif effect == "uncover":
        inner = f'<p:uncover dir="{direction}"/>'
    else:
        return (
            "slide_transition.effect unknown; use fade|cut|push|wipe|split|cover|uncover|random_bar "
            "(push/wipe/cover/uncover use direction; split uses orient)"
        )

    ns_p = "http://schemas.openxmlformats.org/presentationml/2006/main"
    xml = f'<p:transition xmlns:p="{ns_p}" spd="{speed}">{inner}</p:transition>'
    try:
        trans_el = parse_xml(xml)
    except Exception as e:  # noqa: BLE001
        return f"slide_transition XML error: {e!r}"

    el = slide._element
    q_tr = qn("p:transition")
    for child in list(el):
        if child.tag == q_tr:
            el.remove(child)
    anchor = el.clrMapOvr
    if anchor is None:
        anchor = el.cSld
    anchor.addnext(trans_el)
    return None


def _policy_save(path: Path, policy_check: Callable[[Path], None]) -> None:
    path = path.expanduser().resolve()
    policy_check(path)
    path.parent.mkdir(parents=True, exist_ok=True)


def xlsx_write_workbook(
    path: Path,
    sheets: list[dict[str, Any]],
    *,
    policy_check: Callable[[Path], None],
    append: bool = False,
) -> dict[str, Any]:
    """
    写入或追加 Excel。

    ``sheets`` 每项: ``name`` (str), ``headers`` (list, 可选), ``rows`` (list[list]).
    """
    if not isinstance(sheets, list) or not sheets:
        return {"ok": False, "error": "sheets must be non-empty list"}
    if len(sheets) > MAX_SHEETS:
        return {"ok": False, "error": f"max {MAX_SHEETS} sheets"}

    path = path.expanduser().resolve()
    _policy_save(path, policy_check)

    if append and path.is_file():
        wb = load_workbook(path)
    else:
        wb = Workbook()

    used_default_sheet = False

    for si, spec in enumerate(sheets):
        if not isinstance(spec, dict):
            return {"ok": False, "error": "each sheet must be object"}
        name = spec.get("name") or "Sheet"
        if not isinstance(name, str):
            return {"ok": False, "error": "sheet name must be string"}
        name = name[:31] if len(name) > 31 else name
        headers = spec.get("headers")
        rows = spec.get("rows") or []
        if headers is not None and not isinstance(headers, (list, tuple)):
            return {"ok": False, "error": "headers must be list"}
        if not isinstance(rows, list):
            return {"ok": False, "error": "rows must be list"}
        if len(rows) > MAX_ROWS_PER_SHEET:
            return {"ok": False, "error": f"max {MAX_ROWS_PER_SHEET} rows per sheet"}

        if name in wb.sheetnames:
            ws = wb[name]
            mr = ws.max_row or 0
            if mr:
                ws.delete_rows(1, mr)
        elif (
            not used_default_sheet
            and len(wb.sheetnames) == 1
            and wb.active.title == "Sheet"
        ):
            ws = wb.active
            ws.title = name
            used_default_sheet = True
        else:
            ws = wb.create_sheet(title=name)

        if headers:
            if len(headers) > MAX_COLS:
                return {"ok": False, "error": f"max {MAX_COLS} columns"}
            ws.append(list(headers))
        for row in rows:
            if not isinstance(row, (list, tuple)):
                return {"ok": False, "error": "each row must be list"}
            if len(row) > MAX_COLS:
                return {"ok": False, "error": f"max {MAX_COLS} columns"}
            ws.append(list(row))

    wb.save(path)
    return {"ok": True, "path": str(path), "sheet_names": wb.sheetnames}


def pptx_inspect_template(
    path: Path,
    *,
    policy_check: Callable[[Path], None],
) -> dict[str, Any]:
    """列出模板中的版式与占位符，供 AI 选择 layout_index。"""
    path = path.expanduser().resolve()
    policy_check(path)
    if not path.is_file() or path.suffix.lower() != ".pptx":
        return {"ok": False, "error": "path must be .pptx file"}

    prs = Presentation(str(path))
    layouts: list[dict[str, Any]] = []
    for i, layout in enumerate(prs.slide_layouts):
        phs: list[dict[str, Any]] = []
        try:
            for shp in layout.placeholders:
                try:
                    idx = shp.placeholder_format.idx
                    ph_type = int(shp.placeholder_format.type)
                except Exception:
                    idx, ph_type = -1, -1
                phs.append({"idx": idx, "type": ph_type, "name": getattr(shp, "name", "")})
        except Exception:
            pass
        layouts.append({"index": i, "name": layout.name, "placeholders": phs})

    return {
        "ok": True,
        "path": str(path),
        "slide_width_in": prs.slide_width / 914400 * 1.0,
        "slide_height_in": prs.slide_height / 914400 * 1.0,
        "layout_count": len(layouts),
        "layouts": layouts[:40],
        "shape_kind_examples": list(_PPTX_SHAPE_KIND_EXAMPLES),
        "note": "pptx_compose shapes[].kind accepts these examples or any python-pptx MSO_SHAPE name "
        "(e.g. STAR_10_POINT). Per-shape entrance animations are not supported by python-pptx—use "
        "slide_transition and/or a hand-built template with animation on placeholders.",
    }


def pptx_compose(
    path_out: Path,
    slides: list[dict[str, Any]],
    *,
    policy_check: Callable[[Path], None],
    template_path: Path | None = None,
) -> dict[str, Any]:
    """
    从空白或模板生成 PPTX。

    每项 slide: ``layout_index`` (int, 默认 1), ``title`` (str), ``bullets`` (list[str]),
    可选 ``image_path`` (相对工作区或绝对，须通过策略), ``image_left_in``, ``image_top_in``, ``image_width_in``.

    丰富版式（可选，均为英寸）：

    - ``shapes``: 装饰几何，列表项含 ``kind`` (rectangle|rounded_rectangle|oval)、
      ``left_in`` / ``top_in`` / ``width_in`` / ``height_in``，可选 ``fill_rgb`` / ``line_rgb`` 为 ``[R,G,B]``。
      新图形叠在版式占位符之上，请放在页边或留白区，避免挡住标题正文。
    - ``table``: ``data`` 为二维字符串列表；``table_left_in`` / ``table_top_in`` / ``table_width_in`` / ``table_height_in``。
    - ``chart``: ``chart_type`` 含 column/bar/line/area 的 clustered|stacked|stacked_100、doughnut、radar 等；
      ``categories``、``series``（``name`` + ``values``）；**pie / doughnut** 仅允许 **一个** series。
    - ``images``: 多图列表，每项 ``image_path``、``image_left_in``、``image_top_in``、``image_width_in``、
      可选 ``image_height_in``；与单字段 ``image_path`` 可同时使用，合计不超过每页上限。
    - ``text_boxes``: 自由文本框列表（``left_in``…``height_in`` + ``text`` 或 ``lines``），可选 ``font_pt``、``bold``。
    - ``speaker_notes``: 字符串，写入演讲者备注（受长度上限）。
    - ``slide_transition``: 片间切换（非形状动画）— ``effect`` fade|cut|push|wipe|split|cover|uncover|random_bar，
      ``speed`` slow|med|fast，``direction`` l|r|u|d（push/wipe/cover/uncover），``split`` 用 ``orient`` horz|vert。
    - ``shapes[].kind``：除 rectangle/rounded_rectangle/oval 外，可使用 **pptx_inspect_template** 返回的
      ``shape_kind_examples`` 或任意 **MSO_SHAPE** 枚举名（如 ``CHEVRON``、``STAR_5_POINT``）。
    """
    if not isinstance(slides, list) or not slides:
        return {"ok": False, "error": "slides must be non-empty list"}
    if len(slides) > MAX_SLIDES:
        return {"ok": False, "error": f"max {MAX_SLIDES} slides"}

    path_out = path_out.expanduser().resolve()
    _policy_save(path_out, policy_check)

    if template_path is not None:
        tp = template_path.expanduser().resolve()
        policy_check(tp)
        if not tp.is_file() or tp.suffix.lower() != ".pptx":
            return {"ok": False, "error": "template_path must be .pptx"}
        prs = Presentation(str(tp))
    else:
        prs = Presentation()

    for spec in slides:
        if not isinstance(spec, dict):
            return {"ok": False, "error": "each slide must be object"}
        li = int(spec.get("layout_index", 1))
        if li < 0 or li >= len(prs.slide_layouts):
            return {"ok": False, "error": f"layout_index {li} out of range 0..{len(prs.slide_layouts)-1}"}

        slide_layout = prs.slide_layouts[li]
        slide = prs.slides.add_slide(slide_layout)

        title_text = spec.get("title") or ""
        if isinstance(title_text, str) and slide.shapes.title is not None:
            slide.shapes.title.text = title_text

        bullets = spec.get("bullets") or []
        if isinstance(bullets, str):
            bullets = [bullets]
        if not isinstance(bullets, list):
            return {"ok": False, "error": "bullets must be list or string"}
        bullets = [str(b) for b in bullets[:MAX_BULLETS]]

        title_shape = getattr(slide.shapes, "title", None)
        body_shape = None
        try:
            ph_iter = list(slide.placeholders)
        except Exception:
            ph_iter = []
        for shape in ph_iter:
            if not hasattr(shape, "text_frame"):
                continue
            if title_shape is not None and shape == title_shape:
                continue
            body_shape = shape
            break
        if body_shape is None:
            for shape in slide.shapes:
                if title_shape is not None and shape == title_shape:
                    continue
                if hasattr(shape, "text_frame"):
                    body_shape = shape
                    break

        if body_shape is not None:
            tf = body_shape.text_frame
            tf.clear()
            if bullets:
                tf.text = bullets[0]
                for b in bullets[1:]:
                    p = tf.add_paragraph()
                    p.text = b
                    p.level = 0

        err = _pptx_slide_add_shapes(slide, spec.get("shapes"))
        if err:
            return {"ok": False, "error": err}
        err = _pptx_slide_add_table(slide, spec.get("table"))
        if err:
            return {"ok": False, "error": err}
        err = _pptx_slide_add_chart(slide, spec.get("chart"))
        if err:
            return {"ok": False, "error": err}

        pics = _pptx_collect_slide_picture_specs(spec)
        pic_res = _pptx_slide_add_pictures(slide, prs, pics, policy_check)
        if isinstance(pic_res, dict):
            return pic_res
        if isinstance(pic_res, str):
            return {"ok": False, "error": pic_res}

        err = _pptx_slide_add_text_boxes(slide, spec.get("text_boxes"))
        if err:
            return {"ok": False, "error": err}

        err = _pptx_slide_set_speaker_notes(slide, spec.get("speaker_notes"))
        if err:
            return {"ok": False, "error": err}

        err = _pptx_slide_set_transition(slide, spec.get("slide_transition"))
        if err:
            return {"ok": False, "error": err}

    prs.save(path_out)
    return {"ok": True, "path": str(path_out), "slide_count": len(prs.slides)}



"""样式参考：扫描目录/文件，解析 Word 用 YAML/JSON 样式配置。"""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Callable

import yaml

TEXT_EXT = {".md", ".txt", ".markdown", ".css", ".yaml", ".yml", ".json", ".html", ".htm"}
IMAGE_EXT = {".png", ".jpg", ".jpeg", ".gif", ".webp", ".bmp"}


def load_docx_style_profile(
    path: Path,
    *,
    policy_check: Callable[[Path], None],
) -> dict[str, Any]:
    """
    读取 YAML 或 JSON，提取 ``docx_defaults`` 与自由文本 ``notes`` / ``freeform_notes``。

    期望结构示例::

        docx_defaults:
          default_east_asia_font: 宋体
          default_latin_font: Times New Roman
          default_color_hex: "000000"
          default_body_font_pt: 12
        notes: |
          一级标题居中，正文首行缩进两字符（在 blocks 里自行控制）
    """
    path = path.expanduser().resolve()
    policy_check(path)
    if not path.is_file():
        return {"ok": False, "error": f"not a file: {path}"}
    suf = path.suffix.lower()
    if suf not in {".yaml", ".yml", ".json"}:
        return {"ok": False, "error": "style profile must be .yaml, .yml, or .json"}
    raw = path.read_text(encoding="utf-8", errors="replace")
    try:
        if suf == ".json":
            data = json.loads(raw)
        else:
            data = yaml.safe_load(raw)
    except (yaml.YAMLError, json.JSONDecodeError) as e:
        return {"ok": False, "error": f"parse error: {e}"}
    if not isinstance(data, dict):
        return {"ok": False, "error": "root must be object"}

    dd = data.get("docx_defaults")
    if dd is not None and not isinstance(dd, dict):
        return {"ok": False, "error": "docx_defaults must be object"}
    dd = dd or {}

    notes_parts: list[str] = []
    for key in ("notes", "freeform_notes", "description"):
        v = data.get(key)
        if isinstance(v, str) and v.strip():
            notes_parts.append(v.strip())
    notes = "\n\n".join(notes_parts) if notes_parts else ""

    out: dict[str, Any] = {
        "ok": True,
        "profile_path": str(path),
        "docx_defaults": {
            "default_east_asia_font": dd.get("default_east_asia_font"),
            "default_latin_font": dd.get("default_latin_font"),
            "default_color_hex": dd.get("default_color_hex"),
            "default_body_font_pt": dd.get("default_body_font_pt"),
        },
        "notes": notes,
    }
    # drop Nones from docx_defaults for cleaner merge
    out["docx_defaults"] = {k: v for k, v in out["docx_defaults"].items() if v is not None}
    return out


def scan_style_reference(
    root: Path,
    *,
    policy_check: Callable[[Path], None],
    max_files: int = 48,
    max_text_snippet: int = 4000,
    max_total_digest: int = 28000,
) -> dict[str, Any]:
    """
    扫描单个文件或目录，汇总文本类文件片段与图片路径，供模型按用户风格生成内容。
    不调用视觉模型；图片仅列路径与尺寸（Pillow 可读时）。
    """
    root = root.expanduser().resolve()
    policy_check(root)
    entries: list[dict[str, Any]] = []
    digest_parts: list[str] = []

    def _add(kind: str, p: Path, extra: dict[str, Any]) -> None:
        if len(entries) >= max_files:
            return
        e = {"kind": kind, "path": str(p), **extra}
        entries.append(e)
        line = f"[{kind}] {p.name}: {extra.get('summary', '')[:500]}"
        digest_parts.append(line)
        if sum(len(x) for x in digest_parts) > max_total_digest:
            digest_parts.pop()

    def _one_file(p: Path) -> None:
        policy_check(p)
        suf = p.suffix.lower()
        if suf in TEXT_EXT:
            text = p.read_text(encoding="utf-8", errors="replace")
            snip = text[:max_text_snippet]
            if len(text) > max_text_snippet:
                snip += "\n…(truncated)"
            _add("text", p, {"summary": snip, "bytes": len(text.encode("utf-8"))})
        elif suf in IMAGE_EXT:
            w = h = None
            try:
                from PIL import Image

                with Image.open(p) as im:
                    w, h = im.size
            except Exception:
                pass
            _add(
                "image",
                p,
                {"summary": f"image file", "width": w, "height": h},
            )
        elif suf in {".yaml", ".yml", ".json"}:
            pr = load_docx_style_profile(p, policy_check=policy_check)
            if pr.get("ok"):
                _add(
                    "style_profile",
                    p,
                    {
                        "summary": json.dumps(
                            pr.get("docx_defaults", {}), ensure_ascii=False
                        )
                        + (f"\nnotes: {pr.get('notes', '')[:800]}" if pr.get("notes") else ""),
                    },
                )
            else:
                _add("other", p, {"summary": pr.get("error", "unreadable profile")})
        else:
            _add("other", p, {"summary": f"binary or unknown ext {suf!r}"})

    if root.is_file():
        _one_file(root)
    elif root.is_dir():
        for p in sorted(root.rglob("*")):
            if len(entries) >= max_files:
                break
            if p.is_dir():
                continue
            if p.name.startswith("."):
                continue
            rel_depth = len(p.relative_to(root).parts)
            if rel_depth > 4:
                continue
            try:
                _one_file(p)
            except PermissionError:
                continue
    else:
        return {"ok": False, "error": f"not a file or directory: {root}"}

    digest = "\n".join(digest_parts)
    if len(digest) > max_total_digest:
        digest = digest[: max_total_digest - 20] + "\n…(digest truncated)"

    return {
        "ok": True,
        "root": str(root),
        "entry_count": len(entries),
        "entries": entries,
        "style_digest": digest,
        "hint": "Use docx_defaults from style_profile entries with docx_compose; for PPTX use pptx_inspect_template on user template then pptx_compose.",
    }



from __future__ import annotations

import asyncio
from pathlib import Path

from aicd.tools import db_multi


def test_is_readonly_sql() -> None:
    assert db_multi.is_readonly_sql("SELECT 1")
    assert db_multi.is_readonly_sql("WITH x AS (SELECT 1) SELECT * FROM x")
    assert db_multi.is_readonly_sql("SHOW TABLES")
    assert db_multi.is_readonly_sql("DESC foo")
    assert not db_multi.is_readonly_sql("INSERT INTO t VALUES (1)")
    assert not db_multi.is_readonly_sql("UPDATE t SET x=1")
    assert not db_multi.is_readonly_sql("DELETE FROM t")


def test_write_confirm_required(tmp_path: Path) -> None:
    dbf = tmp_path / "w.db"

    async def _go() -> None:
        spec = db_multi.connection_from_dict(
            {"engine": "sqlite", "path": str(dbf.resolve())}
        )
        r = await db_multi.db_execute(
            spec,
            "CREATE TABLE a(id INT)",
            read_only=False,
            max_rows=10,
            write_confirm=None,
        )
        assert r["ok"] is False
        r2 = await db_multi.db_execute(
            spec,
            "CREATE TABLE a(id INT)",
            read_only=False,
            max_rows=10,
            write_confirm=db_multi.WRITE_CONFIRM,
        )
        assert r2.get("ok") is True

    asyncio.run(_go())


def test_sqlite_inspect_and_query(tmp_path: Path) -> None:
    dbf = tmp_path / "x.db"

    async def _go() -> None:
        import aiosqlite

        async with aiosqlite.connect(dbf) as db:
            await db.execute("CREATE TABLE jobs (id INTEGER PRIMARY KEY, name TEXT)")
            await db.execute("INSERT INTO jobs (name) VALUES ('a')")
            await db.commit()

        spec = db_multi.connection_from_dict(
            {"engine": "sqlite", "path": str(dbf.resolve())}
        )
        ping = await db_multi.db_ping(spec)
        assert ping["ok"] is True

        tables = await db_multi.db_inspect_operation(spec, "list_tables")
        assert tables["ok"] is True
        names = [r.get("name") for r in tables.get("rows", [])]
        assert "jobs" in names

        desc = await db_multi.db_inspect_operation(
            spec, "describe_table", table="jobs"
        )
        assert desc["ok"] is True

        q = await db_multi.db_execute(
            spec,
            "SELECT * FROM jobs",
            read_only=True,
            max_rows=50,
            write_confirm=None,
        )
        assert q["ok"] is True
        assert len(q["rows"]) == 1

    asyncio.run(_go())


def test_multi_statement_rejected() -> None:
    async def _go() -> None:
        spec = db_multi.connection_from_dict(
            {"engine": "sqlite", "path": ":memory:"}
        )
        r = await db_multi.db_execute(
            spec,
            "SELECT 1; SELECT 2",
            read_only=True,
            max_rows=10,
            write_confirm=None,
        )
        assert r["ok"] is False

    asyncio.run(_go())



from __future__ import annotations

from pathlib import Path

import pytest

pytest.importorskip("reportlab")

from aicd.tools.pdf_compose_rl import pdf_compose


def test_pdf_compose_minimal(tmp_path: Path) -> None:
    out = tmp_path / "c.pdf"
    r = pdf_compose(
        out,
        [
            {"type": "heading", "level": 1, "text": "Hello"},
            {"type": "paragraph", "text": "Line one.\nLine two."},
        ],
        policy_check=lambda p: None,
    )
    assert r.get("ok") is True
    assert out.is_file() and out.stat().st_size > 80



"""通过 https://embed.diagrams.net（proto=json）加载 .drawio XML 并导出 PNG。"""

from __future__ import annotations

import base64
import json
import os
import re
import subprocess
from pathlib import Path
from typing import Any, Callable

from aicd.kit import drawio_spec

MAX_DRAWIO_XML_CHARS = 3_000_000

# 环境变量（启动进程前设置）：
# - AICD_ENABLE_DRAWIO_EXPORT_PNG=1 — **开启** `.drawio`→PNG 导出（默认关闭；未开启时工具不会注册给模型）
# - AICD_DISABLE_DRAWIO_EXPORT_PNG=1 — 强制关闭（即使已设 ENABLE）
# - AICD_DRAWIO_OFFLINE_ONLY=1 — 禁止 Playwright + embed.diagrams.net；仅允许 AICD_DRAWIO_CLI 本机导出

IFRAME_SRC = (
    "https://embed.diagrams.net/?embed=1&proto=json&spin=1"
    "&noSaveBtn=1&noExitBtn=1&saveAndExit=0&libraries=0"
)


def _parse_data_uri_png(data_uri: str) -> bytes:
    s = (data_uri or "").strip()
    m = re.match(r"^data:image/png;base64,(.+)$", s, re.I | re.DOTALL)
    if not m:
        raise ValueError("export data is not a PNG data URI")
    return base64.b64decode(m.group(1), validate=True)


def _build_parent_page(diagram_xml: str) -> str:
    xml_lit = json.dumps(diagram_xml, ensure_ascii=False)
    return f"""<!DOCTYPE html>
<html lang="zh-CN">
<head>
  <meta charset="utf-8"/>
  <title>drawio embed export</title>
</head>
<body style="margin:0;padding:8px;background:#f0f0f0;">
<iframe
  id="dio"
  title="drawio"
  src="{IFRAME_SRC}"
  style="width:1600px;height:1000px;border:0;display:block;"
></iframe>
<script>
(function () {{
  const DIAGRAM_XML = {xml_lit};
  window.__DRAWIO_DONE__ = false;
  window.__DRAWIO_ERROR__ = null;
  window.__DRAWIO_PNG_URI__ = null;

  function parseMsg(data) {{
    if (data == null) return null;
    if (typeof data === "string") {{
      try {{ return JSON.parse(data); }} catch (e) {{ return null; }}
    }}
    if (typeof data === "object") return data;
    return null;
  }}

  window.addEventListener("message", function (event) {{
    const iframe = document.getElementById("dio");
    if (!iframe || event.source !== iframe.contentWindow) return;
    const msg = parseMsg(event.data);
    if (!msg || typeof msg !== "object") return;

    if (msg.error) {{
      window.__DRAWIO_ERROR__ = JSON.stringify(msg);
      window.__DRAWIO_DONE__ = true;
      return;
    }}

    if (msg.event === "init") {{
      iframe.contentWindow.postMessage(JSON.stringify({{
        action: "load",
        xml: DIAGRAM_XML
      }}), "*");
      return;
    }}

    if (msg.event === "load") {{
      iframe.contentWindow.postMessage(JSON.stringify({{
        action: "export",
        format: "png",
        scale: __EXPORT_SCALE__,
        transparent: __EXPORT_TRANSPARENT__,
        border: __EXPORT_BORDER__,
        size: "diagram"
      }}), "*");
      return;
    }}

    if (msg.event === "export") {{
      if (msg.format === "png" && msg.data) {{
        window.__DRAWIO_PNG_URI__ = msg.data;
        window.__DRAWIO_DONE__ = true;
        return;
      }}
      window.__DRAWIO_ERROR__ = JSON.stringify(msg);
      window.__DRAWIO_DONE__ = true;
    }}
  }});
}})();
</script>
</body>
</html>
"""


def _env_flag(name: str) -> bool:
    return (os.environ.get(name) or "").strip().lower() in ("1", "true", "yes")


def drawio_export_png_allowed() -> bool:
    """
    是否允许使用 ``diagram_drawio_export_png``（注册工具 + 执行导出）。

    默认 **关闭**；需 ``AICD_ENABLE_DRAWIO_EXPORT_PNG=1``。
    ``AICD_DISABLE_DRAWIO_EXPORT_PNG=1`` 时恒为 False。
    """
    if _env_flag("AICD_DISABLE_DRAWIO_EXPORT_PNG"):
        return False
    return _env_flag("AICD_ENABLE_DRAWIO_EXPORT_PNG")


def _export_via_drawio_desktop_cli(
    src: Path,
    out: Path,
    scale: float,
) -> dict[str, Any] | None:
    """
    可选：若设置环境变量 **AICD_DRAWIO_CLI** 为 diagrams.net / draw.io **桌面版可执行文件**路径，
    则优先用其命令行导出（适合**无外网**或 embed 失败时）。未设置则返回 None。
    """
    cli = (os.environ.get("AICD_DRAWIO_CLI") or "").strip()
    if not cli:
        return None
    exe = Path(cli)
    if not exe.is_file():
        return {
            "ok": False,
            "error": f"AICD_DRAWIO_CLI is not an existing file: {cli}",
            "hint": "指向 draw.io / diagrams.net 安装目录下的可执行文件（Windows 常为 draw.io.exe）。",
        }
    sc = max(0.5, min(8.0, float(scale)))
    attempts: list[list[str]] = [
        [str(exe), "-x", "-f", "png", "--scale", str(sc), "-o", str(out), str(src)],
        [str(exe), "-x", "-f", "png", "-o", str(out), str(src)],
    ]
    last_err = ""
    for cmd in attempts:
        try:
            cp = subprocess.run(
                cmd,
                capture_output=True,
                text=True,
                timeout=180,
                check=False,
            )
            if out.is_file() and out.stat().st_size > 0:
                return {
                    "ok": True,
                    "png_path": str(out),
                    "source_drawio": str(src),
                    "scale": sc,
                    "transparent": False,
                    "border": 0,
                    "via": "drawio_desktop_cli",
                }
            last_err = (cp.stderr or cp.stdout or f"exit {cp.returncode}")[:2000]
        except subprocess.TimeoutExpired:
            last_err = "draw.io CLI timed out after 180s"
        except OSError as e:
            last_err = str(e)
    return {
        "ok": False,
        "error": f"draw.io desktop CLI export failed: {last_err}",
        "hint": "确认 AICD_DRAWIO_CLI 与参数兼容；Linux 无桌面时可能需 xvfb-run。",
    }


def _playwright_drawio_hint(exc: BaseException) -> str:
    msg = str(exc)
    parts = [
        "需要可访问 https://embed.diagrams.net 的出站 HTTPS；",
        "以及 Playwright 的 Chromium（执行: python -m playwright install chromium）。",
    ]
    low = msg.lower()
    if "executable doesn't exist" in low or "browserType.launch" in msg:
        parts.insert(
            0,
            "未检测到 Chromium：请先执行 `python -m playwright install chromium`。",
        )
    if "timeout" in low or "navigation" in low:
        parts.append("若网络受限，可安装 draw.io 桌面版并设置环境变量 AICD_DRAWIO_CLI 后重试导出。")
    return " ".join(parts)


def export_drawio_file_to_png(
    drawio_path: Path,
    output_png: Path,
    *,
    policy_check: Callable[[Path], None],
    scale: float = 3.0,
    transparent: bool = False,
    border: int = 2,
    timeout_ms: int = 120_000,
) -> dict[str, Any]:
    """
    导出顺序：

    1. 若设置了 **AICD_DRAWIO_CLI**，先尝试 **draw.io 桌面版**命令行导出（可离线）。
    2. 否则使用 **Playwright 无头 Chromium** 加载本地 HTML，内嵌 iframe 调用 **embed.diagrams.net**
       的 JSON 协议（``load`` → ``export``）生成 PNG（需出站 HTTPS）。

    插入 Word 前请确保本函数返回 **ok** 且 PNG 路径存在，再在 **docx_compose** 的 **image** 块中引用该路径。

    启用 / 禁用 / 离线策略见模块顶部环境变量说明。
    """
    if not drawio_export_png_allowed():
        return {
            "ok": False,
            "error": "diagram_drawio_export_png 未启用（默认关闭）。请设置 AICD_ENABLE_DRAWIO_EXPORT_PNG=1。",
            "hint": "`.drawio` 源文件可用 diagram_drawio_write 离线生成；Word 插图可改用 diagram_mermaid_png，"
            "或在 Draw.io 桌面版手动导出 PNG。若已设 ENABLE 仍失败，检查是否被 AICD_DISABLE_DRAWIO_EXPORT_PNG=1 覆盖。",
        }

    src = drawio_path.expanduser().resolve()
    policy_check(src)
    out = output_png.expanduser().resolve()
    policy_check(out)
    if not src.is_file():
        return {"ok": False, "error": f"drawio file not found: {src}"}
    if out.suffix.lower() != ".png":
        return {"ok": False, "error": "output path must end with .png"}

    raw = src.read_text(encoding="utf-8", errors="replace")
    raw = drawio_spec.strip_bom_and_xml_decl(raw)
    if len(raw) > MAX_DRAWIO_XML_CHARS:
        return {
            "ok": False,
            "error": f"drawio XML exceeds {MAX_DRAWIO_XML_CHARS} chars",
        }
    if "<mxfile" not in raw.lower():
        return {"ok": False, "error": "file does not look like draw.io XML (expected mxfile)"}

    sc = max(0.5, min(8.0, float(scale)))
    bd = max(0, min(64, int(border)))
    tr = bool(transparent)

    offline_only = _env_flag("AICD_DRAWIO_OFFLINE_ONLY")

    cli_first: dict[str, Any] | None = _export_via_drawio_desktop_cli(src, out, sc)
    if cli_first is not None:
        if cli_first.get("ok"):
            return cli_first
        err_t = str(cli_first.get("error") or "")
        if "not an existing file" in err_t.lower():
            return cli_first
        if offline_only:
            hint = str(cli_first.get("hint") or "").strip()
            extra = "已启用 AICD_DRAWIO_OFFLINE_ONLY，不会使用需访问 embed.diagrams.net 的导出方式。"
            return {
                **cli_first,
                "hint": f"{hint} {extra}".strip() if hint else extra,
            }
        # 已安装 CLI 但导出失败时仍尝试 Playwright + embed（便于内网失败时走在线）

    if offline_only:
        return {
            "ok": False,
            "error": "已启用 AICD_DRAWIO_OFFLINE_ONLY，但未通过本机 draw.io 桌面版完成导出。",
            "hint": "将 AICD_DRAWIO_CLI 设为安装目录下的 draw.io 可执行文件（本机路径，非 URL），"
            "或改用 diagram_mermaid_png；不需要该功能时不要设置 AICD_ENABLE_DRAWIO_EXPORT_PNG。",
        }

    try:
        from playwright.sync_api import sync_playwright
    except ImportError as e:
        return {"ok": False, "error": f"playwright not importable: {e}"}

    html = _build_parent_page(raw)
    html = (
        html.replace("__EXPORT_SCALE__", str(sc))
        .replace("__EXPORT_TRANSPARENT__", "true" if tr else "false")
        .replace("__EXPORT_BORDER__", str(bd))
    )

    work = out.parent
    work.mkdir(parents=True, exist_ok=True)
    policy_check(work)
    html_path = work / "_aicd_drawio_export_host.html"
    html_path.write_text(html, encoding="utf-8")
    policy_check(html_path)
    uri = html_path.resolve().as_uri()

    try:
        with sync_playwright() as p:
            browser = p.chromium.launch(headless=True)
            try:
                context = browser.new_context()
                page = context.new_page()
                page.set_default_timeout(timeout_ms)
                page.goto(uri, wait_until="load")
                page.wait_for_function(
                    "() => window.__DRAWIO_DONE__ === true",
                    timeout=timeout_ms,
                )
                err = page.evaluate("() => window.__DRAWIO_ERROR__")
                if err:
                    return {
                        "ok": False,
                        "error": f"draw.io embed/export failed: {err}",
                        "hint": "请检查能否访问 embed.diagrams.net（代理/防火墙）。"
                        "离线可安装 draw.io 桌面版并设置环境变量 AICD_DRAWIO_CLI 指向可执行文件。",
                    }
                data_uri = page.evaluate("() => window.__DRAWIO_PNG_URI__")
                if not data_uri or not isinstance(data_uri, str):
                    return {
                        "ok": False,
                        "error": "no PNG data returned from draw.io export",
                    }
                png_bytes = _parse_data_uri_png(data_uri)
                out.write_bytes(png_bytes)
            finally:
                browser.close()
    except Exception as e:
        out_d: dict[str, Any] = {
            "ok": False,
            "error": str(e),
            "hint": _playwright_drawio_hint(e),
        }
        if (
            cli_first is not None
            and not cli_first.get("ok")
            and (os.environ.get("AICD_DRAWIO_CLI") or "").strip()
        ):
            out_d["cli_attempt_error"] = cli_first.get("error")
        return out_d
    finally:
        try:
            html_path.unlink(missing_ok=True)
        except OSError:
            pass

    if not out.is_file() or out.stat().st_size == 0:
        return {"ok": False, "error": "PNG not written or empty"}

    return {
        "ok": True,
        "png_path": str(out),
        "source_drawio": str(src),
        "scale": sc,
        "transparent": tr,
        "border": bd,
    }



"""大仓库：对已落盘的生成树做过滤、子树、统计，避免整份 JSON 塞进对话。"""

from __future__ import annotations

import json
import re
from collections import defaultdict
from pathlib import Path
from typing import Any


def load_tree_file(path: Path) -> dict[str, Any]:
    if not path.is_file():
        return {"ok": False, "error": "tree file not found"}
    try:
        data = json.loads(path.read_text(encoding="utf-8"))
    except (OSError, json.JSONDecodeError) as e:
        return {"ok": False, "error": str(e)}
    if not isinstance(data, dict):
        return {"ok": False, "error": "invalid tree json"}
    nodes = data.get("nodes")
    if not isinstance(nodes, list):
        return {"ok": False, "error": "missing nodes array"}
    return {"ok": True, "data": data, "nodes": nodes}


def _children_map(nodes: list[dict[str, Any]]) -> dict[str | None, list[dict[str, Any]]]:
    m: dict[str | None, list[dict[str, Any]]] = defaultdict(list)
    for n in nodes:
        pid = n.get("parent_id")
        if pid is not None and not isinstance(pid, str):
            pid = str(pid) if pid is not None else None
        m[pid].append(n)
    for k in m:
        m[k].sort(key=lambda x: (str(x.get("file", "")), int(x.get("line_start") or 0)))
    return m


def _collect_subtree_ids(
    cmap: dict[str | None, list[dict[str, Any]]],
    root_id: str,
    max_depth: int,
) -> set[str]:
    out: set[str] = {root_id}
    frontier = [(root_id, 0)]
    while frontier:
        nid, d = frontier.pop()
        if d >= max_depth:
            continue
        for ch in cmap.get(nid, []):
            cid = ch.get("id")
            if isinstance(cid, str) and cid not in out:
                out.add(cid)
                frontier.append((cid, d + 1))
    return out


def query_tree(
    path: Path,
    *,
    action: str,
    parent_id: str | None = None,
    subtree_depth: int = 8,
    qname_contains: str | None = None,
    file_contains: str | None = None,
    kinds: list[str] | None = None,
    source_langs: list[str] | None = None,
    limit: int = 200,
    offset: int = 0,
    include_docstrings: bool = False,
) -> dict[str, Any]:
    loaded = load_tree_file(path)
    if not loaded.get("ok"):
        return loaded
    data = loaded["data"]
    nodes: list[dict[str, Any]] = loaded["nodes"]
    meta = {
        "index_revision": data.get("index_revision"),
        "topic_slug": data.get("topic_slug"),
        "schema": data.get("schema"),
        "scan_root": data.get("root"),
        "files_scanned": data.get("files_scanned"),
        "total_nodes": len(nodes),
        "tree_path": str(path.resolve()),
    }

    if action == "stats":
        by_kind: dict[str, int] = defaultdict(int)
        by_lang: dict[str, int] = defaultdict(int)
        by_file: dict[str, int] = defaultdict(int)
        for n in nodes:
            by_kind[str(n.get("kind") or "?")] += 1
            by_lang[str(n.get("source_lang") or "python")] += 1
            fp = str(n.get("file") or "")
            by_file[fp] += 1
        top_files = sorted(by_file.items(), key=lambda x: -x[1])[:40]
        return {
            "ok": True,
            "action": "stats",
            "meta": meta,
            "by_kind": dict(by_kind),
            "by_source_lang": dict(by_lang),
            "distinct_files": len(by_file),
            "top_files_by_node_count": [{"file": a, "nodes": b} for a, b in top_files],
        }

    cmap = _children_map(nodes)
    kind_set = {k.lower() for k in kinds} if kinds else None
    lang_set = {k.lower() for k in source_langs} if source_langs else None
    qn_sub = (qname_contains or "").lower()
    fc_sub = (file_contains or "").lower()

    def _filter_base(seq: list[dict[str, Any]]) -> list[dict[str, Any]]:
        out: list[dict[str, Any]] = []
        for n in seq:
            if kind_set and str(n.get("kind") or "").lower() not in kind_set:
                continue
            if lang_set:
                sl = str(n.get("source_lang") or "python").lower()
                if sl not in lang_set:
                    continue
            if qn_sub and qn_sub not in str(n.get("qname") or "").lower():
                continue
            if fc_sub and fc_sub not in str(n.get("file") or "").lower():
                continue
            out.append(n)
        return out

    def _shrink(n: dict[str, Any]) -> dict[str, Any]:
        keys = [
            "id",
            "parent_id",
            "kind",
            "name",
            "qname",
            "file",
            "line_start",
            "line_end",
            "source_lang",
            "parser",
        ]
        r = {k: n.get(k) for k in keys}
        if include_docstrings:
            r["docstring_preview"] = n.get("docstring_preview")
            r["ai_description_slot"] = n.get("ai_description_slot")
        return r

    if action == "children":
        if not parent_id:
            return {"ok": False, "error": "children action requires parent_id"}
        raw = cmap.get(parent_id, [])
        filt = _filter_base(raw)
        slice_f = filt[offset : offset + max(1, min(500, limit))]
        return {
            "ok": True,
            "action": "children",
            "meta": meta,
            "parent_id": parent_id,
            "returned": len(slice_f),
            "total_matched": len(filt),
            "nodes": [_shrink(x) for x in slice_f],
        }

    if action == "subtree":
        if not parent_id:
            return {"ok": False, "error": "subtree action requires parent_id"}
        depth = max(1, min(64, subtree_depth))
        ids = _collect_subtree_ids(cmap, parent_id, depth)
        sub = [n for n in nodes if n.get("id") in ids]
        filt = _filter_base(sub)
        slice_f = filt[offset : offset + max(1, min(500, limit))]
        return {
            "ok": True,
            "action": "subtree",
            "meta": meta,
            "root_id": parent_id,
            "subtree_depth_limit": depth,
            "returned": len(slice_f),
            "total_matched": len(filt),
            "nodes": [_shrink(x) for x in slice_f],
        }

    if action == "filter":
        filt = _filter_base(nodes)
        slice_f = filt[offset : offset + max(1, min(500, limit))]
        return {
            "ok": True,
            "action": "filter",
            "meta": meta,
            "returned": len(slice_f),
            "total_matched": len(filt),
            "offset": offset,
            "limit": limit,
            "nodes": [_shrink(x) for x in slice_f],
        }

    return {"ok": False, "error": f"unknown action: {action}"}


def query_tree_regex(
    path: Path,
    *,
    qname_regex: str,
    limit: int = 100,
) -> dict[str, Any]:
    try:
        rx = re.compile(qname_regex, re.I)
    except re.error as e:
        return {"ok": False, "error": f"invalid regex: {e}"}
    loaded = load_tree_file(path)
    if not loaded.get("ok"):
        return loaded
    data = loaded["data"]
    nodes = loaded["nodes"]
    meta = {
        "index_revision": data.get("index_revision"),
        "topic_slug": data.get("topic_slug"),
        "total_nodes": len(nodes),
        "tree_path": str(path.resolve()),
    }
    out = [n for n in nodes if rx.search(str(n.get("qname") or ""))]
    out = out[: max(1, min(300, limit))]
    return {
        "ok": True,
        "action": "regex",
        "meta": meta,
        "returned": len(out),
        "nodes": [
            {
                "id": n.get("id"),
                "parent_id": n.get("parent_id"),
                "kind": n.get("kind"),
                "qname": n.get("qname"),
                "file": n.get("file"),
                "line_start": n.get("line_start"),
                "line_end": n.get("line_end"),
            }
            for n in out
        ],
    }



"""将本地离线 HTML（如 viz_export_html / viz_data_chart）渲染为 PNG，不删除源 HTML。"""

from __future__ import annotations

from pathlib import Path
from typing import Any, Callable


def render_html_file_to_png(
    html_path: Path,
    output_png: Path,
    *,
    policy_check: Callable[[Path], None],
    selector: str | None = None,
    full_page: bool = True,
    device_scale_factor: float = 2.0,
    wait_after_load_ms: int = 1200,
    timeout_ms: int = 120_000,
) -> dict[str, Any]:
    """
    用 Playwright 打开 ``file://`` 页面并截图。

    - **不修改、不删除** ``html_path``；同级 ``res/`` 由浏览器按相对路径加载。
    - ``selector`` 非空时截取该元素，否则整页（``full_page`` 控制是否完整滚动高度）。
    """
    hp = html_path.expanduser().resolve()
    out = output_png.expanduser().resolve()
    policy_check(hp)
    policy_check(out)
    if not hp.is_file():
        return {"ok": False, "error": f"html file not found: {hp}"}
    if hp.suffix.lower() not in {".html", ".htm"}:
        return {"ok": False, "error": "html_path must be .html or .htm"}
    if out.suffix.lower() != ".png":
        return {"ok": False, "error": "output_png_path must end with .png"}

    out.parent.mkdir(parents=True, exist_ok=True)
    uri = hp.resolve().as_uri()

    try:
        from playwright.sync_api import sync_playwright
    except ImportError as e:
        return {
            "ok": False,
            "error": f"playwright not available: {e}",
            "hint": "pip install playwright && python -m playwright install chromium",
        }

    sel = (selector or "").strip() or None
    dsf = max(1.0, min(4.0, float(device_scale_factor)))
    wait_ms = max(0, min(60_000, int(wait_after_load_ms)))

    try:
        with sync_playwright() as p:
            browser = p.chromium.launch(headless=True)
            try:
                context = browser.new_context(device_scale_factor=dsf)
                page = context.new_page()
                page.set_default_timeout(timeout_ms)
                page.goto(uri, wait_until="load")
                if wait_ms:
                    page.wait_for_timeout(wait_ms)
                if sel:
                    loc = page.locator(sel).first
                    loc.wait_for(state="visible", timeout=timeout_ms)
                    loc.screenshot(path=str(out), type="png")
                else:
                    page.screenshot(
                        path=str(out),
                        full_page=bool(full_page),
                        type="png",
                    )
            finally:
                browser.close()
    except Exception as e:
        msg = str(e).lower()
        hint = "运行: python -m playwright install chromium"
        if "executable doesn't exist" in msg or "browsertype.launch" in msg:
            hint = "未找到 Chromium，请执行: python -m playwright install chromium"
        return {"ok": False, "error": str(e), "hint": hint}

    if not out.is_file() or out.stat().st_size == 0:
        return {"ok": False, "error": "PNG was not written or is empty"}

    return {
        "ok": True,
        "png_path": str(out),
        "html_path": str(hp),
        "source_retained": True,
        "note": "Source .html and sibling res/ are unchanged; use png_path in docx_compose / pptx_compose.",
    }



from __future__ import annotations

import asyncio
import base64
import io
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

from PIL import Image


def test_vision_analyze_image_tool_mock(tmp_path: Path) -> None:
    from aicd.runtime import build_runtime

    async def _go() -> None:
        ctx = await build_runtime(home=tmp_path / "home")
        ok, _ = ctx.set_ai_workspace(str(tmp_path))
        assert ok is True

        img = tmp_path / "plot.png"
        Image.new("RGB", (32, 24), color=(40, 80, 120)).save(img)
        img_res = str(img.resolve())

        mock_llm = MagicMock()
        mock_llm.chat_with_images = AsyncMock(
            return_value={
                "analysis": "Trend: sales increase Q1–Q3.",
                "per_image": [
                    {
                        "path": img_res,
                        "source_width": 32,
                        "source_height": 24,
                        "sent_width": 32,
                        "sent_height": 24,
                    }
                ],
            }
        )
        ctx.router.set_vision_llm(mock_llm)

        r = await ctx.router.run(
            "vision_analyze_image",
            {
                "prompt": "Describe the chart trend.",
                "image_path": str(img),
                "save_report": True,
            },
        )
        assert r.get("ok") is True
        assert "Q1" in r.get("analysis", "")
        assert isinstance(r.get("per_image"), list) and len(r["per_image"]) == 1
        assert "原始文件" in (r.get("word_insert_note") or "")
        assert r.get("report_path")
        rp = Path(r["report_path"])
        assert rp.is_file()
        assert "Trend:" in rp.read_text(encoding="utf-8")
        mock_llm.chat_with_images.assert_awaited_once()
        await ctx.conn.close()

    asyncio.run(_go())


def test_image_file_to_data_url_png(tmp_path: Path) -> None:
    from aicd.agent.llm import image_file_to_data_url

    img = tmp_path / "a.png"
    Image.new("RGBA", (4, 4), color=(255, 0, 0, 128)).save(img)
    url = image_file_to_data_url(img)
    assert url.startswith("data:image/png;base64,")
    assert len(url) > 100


def test_image_file_to_data_url_long_edge_1920(tmp_path: Path) -> None:
    from aicd.agent.llm import image_file_to_data_url

    img = tmp_path / "wide.png"
    Image.new("RGB", (2600, 900), color=(10, 20, 30)).save(img)
    url = image_file_to_data_url(img)
    raw_png = base64.standard_b64decode(url.split(",", 1)[1])
    out = Image.open(io.BytesIO(raw_png))
    assert max(out.size) <= 1920


def test_image_file_to_data_url_large_source_compressed(tmp_path: Path) -> None:
    from aicd.agent.llm import image_file_to_data_url

    img = tmp_path / "noisy.png"
    # 噪声图 PNG 往往 > 1MiB，触发「源 ≥1M 压缩」与编码后 ≤1M 迭代
    Image.effect_noise((1800, 1800), 12).convert("RGB").save(img)
    assert img.stat().st_size >= 1024 * 1024
    url = image_file_to_data_url(img)
    raw_png = base64.standard_b64decode(url.split(",", 1)[1])
    assert len(raw_png) <= 1024 * 1024
    out = Image.open(io.BytesIO(raw_png))
    assert max(out.size) <= 1920



"""图谱 / Chart.js / Mermaid / 可视化 HTML 结构检查与轻量清洗建议。"""

from __future__ import annotations

import json
import re
from typing import Any

MAX_MERMAID_CHARS = 200_000
MAX_HTML_CHARS = 600_000
MAX_GRAPH_NODES = 20_000
MAX_GRAPH_EDGES = 100_000

_MAX_NETWORK_REF_ISSUES = 24


def _is_network_url(url: str) -> bool:
    """True if URL loads over the network (http(s) or protocol-relative //)."""
    u = (url or "").strip()
    if not u:
        return False
    # srcset: "https://x 1x" → first token
    u = u.split()[0]
    ul = u.lower()
    if ul.startswith("#"):
        return False
    if ul.startswith(
        ("data:", "mailto:", "tel:", "javascript:", "about:", "blob:")
    ):
        return False
    return ul.startswith("http://") or ul.startswith("https://") or ul.startswith("//")


def _network_ref_issues_from_html(html: str) -> list[str]:
    """
    Exported dashboard HTML must not reference remote assets (offline + file:// safe).
    Detects common tag attributes and url(...) in inline / <style> CSS.
    """
    issues: list[str] = []
    seen: set[str] = set()

    def push(where: str, raw: str) -> None:
        if len(issues) >= _MAX_NETWORK_REF_ISSUES:
            return
        key = (where, raw.strip()[:512])
        if key in seen:
            return
        seen.add(key)
        preview = raw.strip()[:100] + ("…" if len(raw.strip()) > 100 else "")
        issues.append(
            f"network_ref: {where} uses off-page URL ({preview}) — "
            "forbidden: do not use http(s):// or // in viz HTML; download into "
            "sibling **res/** or **deliverables/_assets/** and use a relative path "
            "(see **scripts/download_viz_res.py** for bundled Chart/Mermaid/vis)."
        )

    def check_attr(where: str, value: str | None) -> None:
        if not value:
            return
        for chunk in re.split(r",\s*", value):
            tok = chunk.strip().split()[0] if chunk.strip() else ""
            if _is_network_url(tok):
                push(where, tok)

    _TAG_ATTR = (
        (r"""<\s*script[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<script src>"),
        (r"""<\s*link[^>]*\shref\s*=\s*(["'])(.*?)\1""", "<link href>"),
        (r"""<\s*img[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<img src>"),
        (r"""<\s*iframe[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<iframe src>"),
        (r"""<\s*embed[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<embed src>"),
        (r"""<\s*object[^>]*\sdata\s*=\s*(["'])(.*?)\1""", "<object data>"),
        (r"""<\s*video[^>]*\sposter\s*=\s*(["'])(.*?)\1""", "<video poster>"),
        (r"""<\s*video[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<video src>"),
        (r"""<\s*audio[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<audio src>"),
        (r"""<\s*source[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<source src>"),
        (r"""<\s*track[^>]*\ssrc\s*=\s*(["'])(.*?)\1""", "<track src>"),
        (r"""<\s*base[^>]*\shref\s*=\s*(["'])(.*?)\1""", "<base href>"),
    )
    for pattern, label in _TAG_ATTR:
        for m in re.finditer(pattern, html, re.I | re.DOTALL):
            check_attr(label, m.group(2))

    for m in re.finditer(
        r"""<\s*img[^>]*\ssrcset\s*=\s*(["'])(.*?)\1""",
        html,
        re.I | re.DOTALL,
    ):
        check_attr("<img srcset>", m.group(2))

    for m in re.finditer(
        r"""<\s*source[^>]*\ssrcset\s*=\s*(["'])(.*?)\1""",
        html,
        re.I | re.DOTALL,
    ):
        check_attr("<source srcset>", m.group(2))

    _URL_IN_CSS = re.compile(
        r"""url\(\s*["']?(https?://[^\"')]+|//[^\"')]+)""",
        re.I,
    )
    for m in re.finditer(r"<style[^>]*>(.*?)</style>", html, re.I | re.DOTALL):
        block = m.group(1) or ""
        for um in _URL_IN_CSS.finditer(block):
            if _is_network_url(um.group(1)):
                push("CSS url() in <style>", um.group(1))

    for m in re.finditer(
        r"""style\s*=\s*(["'])(.*?)\1""",
        html,
        re.I | re.DOTALL,
    ):
        for um in _URL_IN_CSS.finditer(m.group(2) or ""):
            if _is_network_url(um.group(1)):
                push("CSS url() in style attribute", um.group(1))

    return issues


def _as_str_id(x: Any) -> str:
    if x is None:
        return ""
    if isinstance(x, bool):
        return str(x).lower()
    return str(x)


def validate_graph_json(
    nodes: list[Any], edges: list[Any], *, normalize: bool = True
) -> dict[str, Any]:
    issues: list[str] = []
    if not isinstance(nodes, list):
        return {"ok": False, "error": "nodes must be a JSON array"}
    if not isinstance(edges, list):
        return {"ok": False, "error": "edges must be a JSON array"}
    if len(nodes) > MAX_GRAPH_NODES:
        issues.append(f"node count {len(nodes)} exceeds cap {MAX_GRAPH_NODES}")
    if len(edges) > MAX_GRAPH_EDGES:
        issues.append(f"edge count {len(edges)} exceeds cap {MAX_GRAPH_EDGES}")

    seen_ids: set[str] = set()
    norm_nodes: list[dict[str, Any]] = []
    for i, n in enumerate(nodes):
        if not isinstance(n, dict):
            issues.append(f"node[{i}] is not an object")
            continue
        nid = n.get("id")
        if nid is None:
            issues.append(f"node[{i}] missing id")
            continue
        sid = _as_str_id(nid)
        if sid in seen_ids:
            issues.append(f"duplicate node id: {sid!r}")
        seen_ids.add(sid)
        lbl = n.get("label", n.get("title", sid))
        entry = {"id": sid, "label": str(lbl) if lbl is not None else sid}
        for k in ("title", "group", "shape", "color"):
            if k in n and n[k] is not None:
                entry[k] = n[k]
        norm_nodes.append(entry)

    norm_edges: list[dict[str, Any]] = []
    for i, e in enumerate(edges):
        if not isinstance(e, dict):
            issues.append(f"edge[{i}] is not an object")
            continue
        a = e.get("from", e.get("source"))
        b = e.get("to", e.get("target"))
        if a is None or b is None:
            issues.append(
                f"edge[{i}] needs from/to or source/target (got keys {list(e.keys())})"
            )
            continue
        sa, sb = _as_str_id(a), _as_str_id(b)
        if sa not in seen_ids:
            issues.append(f"edge[{i}] from/source {sa!r} not in nodes")
        if sb not in seen_ids:
            issues.append(f"edge[{i}] to/target {sb!r} not in nodes")
        ne = {"from": sa, "to": sb}
        if e.get("label") is not None:
            ne["label"] = e["label"]
        if "arrows" in e:
            ne["arrows"] = e["arrows"]
        norm_edges.append(ne)

    ok = not any(
        "missing id" in x
        or "needs from/to" in x
        or "not in nodes" in x
        or "is not an object" in x
        or "duplicate node id" in x
        for x in issues
    )
    out: dict[str, Any] = {
        "ok": ok,
        "issues": issues[:200],
        "issues_truncated": len(issues) > 200,
        "node_count": len(nodes),
        "edge_count": len(edges),
        "unique_node_ids": len(seen_ids),
    }
    if normalize and ok:
        out["normalized"] = {"nodes": norm_nodes, "edges": norm_edges}
    return out


def validate_chartjs_config(cfg: Any) -> dict[str, Any]:
    issues: list[str] = []
    if not isinstance(cfg, dict):
        return {"ok": False, "error": "config must be a JSON object"}
    ct = str(cfg.get("type") or "").strip().lower()
    if ct not in ("bar", "line", "pie", "doughnut", "radar", "scatter", "bubble"):
        if ct:
            issues.append(
                f"type {ct!r} may be unsupported in Aicd helpers (known: bar|line|…)"
            )
        else:
            issues.append("missing type")
    data = cfg.get("data")
    if not isinstance(data, dict):
        issues.append("data must be an object with labels and datasets")
        return {
            "ok": False,
            "issues": issues,
            "hint": "Use chart_spec.series_to_chartjs_config shape",
        }
    labels = data.get("labels")
    dss = data.get("datasets")
    if not isinstance(labels, list):
        issues.append("data.labels must be an array")
    if not isinstance(dss, list) or not dss:
        issues.append("data.datasets must be a non-empty array")
    else:
        ds0 = dss[0]
        if isinstance(ds0, dict) and isinstance(labels, list):
            vals = ds0.get("data")
            if not isinstance(vals, list):
                issues.append("datasets[0].data must be an array of numbers")
            elif len(labels) != len(vals):
                issues.append(
                    f"labels length {len(labels)} != datasets[0].data length {len(vals)}"
                )
        for j, ds in enumerate(dss):
            if not isinstance(ds, dict):
                issues.append(f"datasets[{j}] must be object")
    hard = [
        x
        for x in issues
        if "may be unsupported" not in x and "unsupported in Aicd" not in x
    ]
    ok = len(hard) == 0
    return {"ok": ok, "issues": issues, "type": ct or None}


def validate_mermaid_body(body: str) -> dict[str, Any]:
    if not isinstance(body, str):
        return {"ok": False, "error": "mermaid_body must be string"}
    s = body.strip()
    if len(s) > MAX_MERMAID_CHARS:
        return {"ok": False, "error": f"body exceeds {MAX_MERMAID_CHARS} chars"}
    issues: list[str] = []
    if not s:
        issues.append("empty diagram")
    if "```" in s:
        issues.append("do not embed triple-backticks inside mermaid_body (tool adds fence)")
    low = s.lower()
    starters = (
        "flowchart",
        "graph",
        "sequencediagram",
        "classdiagram",
        "statediagram",
        "erdiagram",
        "journey",
        "gantt",
        "pie",
        "mindmap",
        "timeline",
        "block-beta",
    )
    if s and not any(low.startswith(x) or f"\n{x}" in low for x in starters):
        issues.append(
            "no common diagram keyword at start (e.g. flowchart LR, sequenceDiagram)"
        )
    ok = bool(s) and "```" not in s
    return {
        "ok": ok,
        "issues": issues,
        "char_count": len(s),
        "preview_head": s[:240].replace("\n", " ") + ("…" if len(s) > 240 else ""),
    }


def validate_html_viz_embed(html: str) -> dict[str, Any]:
    if not isinstance(html, str):
        return {"ok": False, "error": "html must be string"}
    if len(html) > MAX_HTML_CHARS:
        return {"ok": False, "error": f"html exceeds {MAX_HTML_CHARS} chars"}
    lower = html.lower()
    issues: list[str] = []

    def first_idx(sub: str) -> int:
        return lower.find(sub)

    chart_lib = first_idx("chart.umd") >= 0 or first_idx("chart.min.js") >= 0
    mermaid_lib = first_idx("mermaid.min.js") >= 0
    vis_lib = first_idx("vis-network.min.js") >= 0

    inline_scripts = list(re.finditer(r"<script(?![^>]*\ssrc=)[^>]*>", lower))
    first_inline = inline_scripts[0].start() if inline_scripts else -1

    def use_before_lib(needle: str, lib_marker: str) -> bool:
        u = lower.find(needle)
        if u < 0:
            return False
        L = lower.find(lib_marker)
        if L < 0:
            return False
        return u < L and (first_inline < 0 or u < first_inline or u < L)

    if first_inline >= 0:
        if "new chart(" in lower and chart_lib:
            nc = lower.find("new chart(")
            cl = lower.find("chart.umd")
            if cl < 0:
                cl = lower.find("chart.min.js")
            if nc >= 0 and cl >= 0 and nc < cl:
                issues.append(
                    "possible new Chart() before Chart.js script loads (visitors get Chart is not defined)"
                )
        if "vis.network" in lower and vis_lib:
            vn = lower.find("vis.network")
            vl = lower.find("vis-network.min.js")
            if vn >= 0 and vl >= 0 and vn < vl:
                issues.append(
                    "possible vis.Network before vis-network.min.js loads"
                )
        if "mermaid." in lower and mermaid_lib:
            mi = lower.find("mermaid.init")
            if mi < 0:
                mi = lower.find("mermaid.run")
            ml = lower.find("mermaid.min.js")
            if mi >= 0 and ml >= 0 and mi < ml:
                issues.append("Mermaid API used before mermaid.min.js loads")

    if not chart_lib and "new chart(" in lower:
        issues.append("uses new Chart but no local chart script reference found")
    if not vis_lib and "vis.network" in lower:
        issues.append("uses vis.Network but no vis-network.min.js reference found")
    if not mermaid_lib and re.search(r"class=[\"']mermaid[\"']", lower):
        issues.append("mermaid class on element but no mermaid.min.js reference found")

    issues.extend(_network_ref_issues_from_html(html))

    _hard = [
        i
        for i in issues
        if "possible" in i
        or "uses " in i
        or i.startswith("network_ref:")
    ]
    return {
        "ok": len(_hard) == 0,
        "issues": issues,
        "has_chart_js_ref": chart_lib,
        "has_mermaid_ref": mermaid_lib,
        "has_vis_ref": vis_lib,
        "inline_script_count": len(inline_scripts),
    }


def run_operation(operation: str, payload: dict[str, Any]) -> dict[str, Any]:
    op = operation.strip().lower().replace("-", "_")
    if op in ("graph_json", "diagram_graph_json"):
        return validate_graph_json(
            list(payload.get("nodes") or []),
            list(payload.get("edges") or []),
            normalize=bool(payload.get("normalize", True)),
        )
    if op in ("chartjs", "chartjs_config", "chart"):
        cfg = payload.get("config")
        if cfg is None and payload.get("config_json"):
            try:
                cfg = json.loads(str(payload["config_json"]))
            except json.JSONDecodeError as e:
                return {"ok": False, "error": f"config_json: {e}"}
        return validate_chartjs_config(cfg)
    if op in ("mermaid", "mermaid_body"):
        return validate_mermaid_body(str(payload.get("mermaid_body") or payload.get("text") or ""))
    if op in ("html_viz", "html_embed", "html"):
        return validate_html_viz_embed(str(payload.get("html") or payload.get("text") or ""))
    return {
        "ok": False,
        "error": f"unknown operation: {operation!r}",
        "hint": "graph_json|chartjs_config|mermaid_body|html_viz",
    }



"""Mermaid → PNG（Playwright）；无浏览器时跳过。"""

from __future__ import annotations

from pathlib import Path

import pytest

from aicd.config import PathPolicyConfig
from aicd.tools.mermaid_png import render_mermaid_to_png
from aicd.tools.path_policy import PathPolicy


def _chromium_available() -> bool:
    try:
        from playwright.sync_api import sync_playwright

        with sync_playwright() as p:
            browser = p.chromium.launch(headless=True)
            browser.close()
        return True
    except Exception:
        return False


@pytest.mark.skipif(not _chromium_available(), reason="Playwright Chromium not installed")
def test_render_mermaid_flowchart_png(tmp_path: Path) -> None:
    pol = PathPolicy(PathPolicyConfig(full_disk=True))
    out = tmp_path / "flow.png"
    r = render_mermaid_to_png(
        "flowchart TB\n  A[分析] --> B[实现]",
        out,
        policy_check=pol.check_allowed,
        theme="neutral",
    )
    assert r["ok"] is True
    assert out.is_file()
    assert out.stat().st_size > 500



from __future__ import annotations

import math
import statistics
from typing import Any


def _as_floats(values: list[Any]) -> list[float]:
    out: list[float] = []
    for v in values:
        out.append(float(v))
    return out


def describe(values: list[Any]) -> dict[str, Any]:
    """n, min, max, mean, median, stdev (sample, n>=2)."""
    xs = _as_floats(values)
    n = len(xs)
    if n == 0:
        return {"ok": False, "error": "empty values"}
    out: dict[str, Any] = {
        "ok": True,
        "n": n,
        "min": min(xs),
        "max": max(xs),
        "mean": statistics.fmean(xs),
        "median": statistics.median(xs),
    }
    if n >= 2:
        out["stdev"] = statistics.stdev(xs)
    else:
        out["stdev"] = None
    return out


def percentile(values: list[Any], p: float) -> dict[str, Any]:
    """Linear interpolation percentile p in [0,100]."""
    if not 0 <= p <= 100:
        return {"ok": False, "error": "p must be in [0,100]"}
    xs = sorted(_as_floats(values))
    n = len(xs)
    if n == 0:
        return {"ok": False, "error": "empty values"}
    if n == 1:
        return {"ok": True, "percentile": p, "value": xs[0]}
    k = (p / 100) * (n - 1)
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return {"ok": True, "percentile": p, "value": xs[int(k)]}
    d0 = xs[f] * (c - k)
    d1 = xs[c] * (k - f)
    return {"ok": True, "percentile": p, "value": d0 + d1}


def histogram(values: list[Any], bins: int = 10) -> dict[str, Any]:
    """Equal-width histogram; bins >= 1."""
    xs = _as_floats(values)
    n = len(xs)
    if n == 0:
        return {"ok": False, "error": "empty values"}
    b = max(1, int(bins))
    lo, hi = min(xs), max(xs)
    if lo == hi:
        return {
            "ok": True,
            "bin_low": [lo],
            "bin_high": [hi],
            "counts": [n],
            "n_bins": 1,
        }
    width = (hi - lo) / b
    counts = [0] * b
    lows = [lo + i * width for i in range(b)]
    highs = [lo + (i + 1) * width for i in range(b)]
    highs[-1] = hi
    for x in xs:
        if x >= hi:
            counts[-1] += 1
            continue
        idx = min(int((x - lo) / width), b - 1)
        counts[idx] += 1
    return {
        "ok": True,
        "bin_low": lows,
        "bin_high": highs,
        "counts": counts,
        "n_bins": b,
    }


def correlation_pearson(x: list[Any], y: list[Any]) -> dict[str, Any]:
    """Pearson r; requires equal length >= 2."""
    a = _as_floats(x)
    b = _as_floats(y)
    if len(a) != len(b):
        return {"ok": False, "error": "x and y length mismatch"}
    if len(a) < 2:
        return {"ok": False, "error": "need at least 2 pairs"}
    try:
        r = statistics.correlation(a, b)
    except statistics.StatisticsError as e:
        return {"ok": False, "error": str(e)}
    return {"ok": True, "n": len(a), "pearson_r": r}


def run_operation(operation: str, payload: dict[str, Any]) -> dict[str, Any]:
    op = operation.strip().lower()
    if op == "describe":
        return describe(payload.get("values") or [])
    if op == "percentile":
        return percentile(payload.get("values") or [], float(payload.get("p", 50)))
    if op == "histogram":
        return histogram(payload.get("values") or [], int(payload.get("bins", 10)))
    if op == "correlation":
        return correlation_pearson(payload.get("x") or [], payload.get("y") or [])
    return {"ok": False, "error": f"unknown operation: {operation}"}



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aicd.kit import script_syntax, validate_viz

if TYPE_CHECKING:
    from aicd.agent.tool_router import ToolRouter


def _viz_spec_validate(_router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    op = str(args.get("operation") or "")
    payload = {k: v for k, v in args.items() if k != "operation"}
    return validate_viz.run_operation(op, payload)


def _script_syntax_check(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    kind = str(args.get("kind") or args.get("language") or "").strip()
    if not kind:
        return {"ok": False, "error": "kind (or language) is required, e.g. py|js|bash|ps1|bat"}
    path_str = args.get("path")
    source = args.get("source")
    if path_str is not None and str(path_str).strip():
        p = router._resolve_path(str(path_str))
        return script_syntax.run_check(kind, source=None, file_path=p)
    if source is None:
        return {"ok": False, "error": "provide source (string) or path (under workspace policy)"}
    return script_syntax.run_check(kind, source=str(source), file_path=None)


HANDLERS: dict[str, Any] = {
    "viz_spec_validate": _viz_spec_validate,
    "script_syntax_check": _script_syntax_check,
}



"""OpenAI 式消息链校验与裁剪（避免 tool_calls / tool 错位导致 API 400）。"""

from __future__ import annotations

from aicd.agent.chat_context import (
    longest_valid_openai_suffix,
    openai_message_list_valid,
    repair_openai_chat_messages,
)


def _a(tc_ids: list[str]) -> dict:
    return {
        "role": "assistant",
        "content": "",
        "tool_calls": [
            {"id": tid, "type": "function", "function": {"name": "x", "arguments": "{}"}}
            for tid in tc_ids
        ],
    }


def _t(tid: str) -> dict:
    return {"role": "tool", "tool_call_id": tid, "content": "{}"}


def test_valid_user_assistant_tool_chain() -> None:
    msgs = [
        {"role": "user", "content": "u"},
        _a(["1"]),
        _t("1"),
        {"role": "assistant", "content": "done"},
    ]
    assert openai_message_list_valid(msgs)


def test_invalid_orphan_tool_prefix() -> None:
    msgs = [_t("1"), {"role": "user", "content": "u"}]
    assert not openai_message_list_valid(msgs)


def test_invalid_assistant_tc_without_tools() -> None:
    msgs = [{"role": "user", "content": "u"}, _a(["1", "2"]), _t("1")]
    assert not openai_message_list_valid(msgs)


def test_repair_drops_incomplete_tool_round() -> None:
    msgs = [{"role": "user", "content": "u"}, _a(["a", "b"]), _t("a")]
    repair_openai_chat_messages(msgs)
    assert msgs == [{"role": "user", "content": "u"}]


def test_longest_suffix_keeps_complete_round() -> None:
    body = [
        {"role": "user", "content": "old"},
        _a(["x"]),
        _t("x"),
        {"role": "user", "content": "new"},
        _a(["y"]),
        _t("y"),
    ]
    # 盲目 tail[-3:] 会得到 [user new, assistant(y), tool(y)] — 合法
    # 若窗口切断 assistant(x) 与 tool(x)，则非法；最长合法后缀应含完整 x 轮或仅 new 轮
    su = longest_valid_openai_suffix(body, 3)
    assert openai_message_list_valid(su)
    assert len(su) == 3
    assert su[0]["content"] == "new"


def test_longest_suffix_avoids_split_tool_round() -> None:
    """模拟子 Agent：尾部为 assistant(两工具) + 仅 1 条 tool；max_len=2 时不能取这 2 条。"""
    body = [
        {"role": "user", "content": "u"},
        _a(["p1", "p2"]),
        _t("p1"),
    ]
    su = longest_valid_openai_suffix(body, 2)
    assert openai_message_list_valid(su)
    assert len(su) == 1
    assert su[0]["role"] == "user"




from __future__ import annotations

import fnmatch
import os
import re
from pathlib import Path
from typing import Any

from aicd.config import PathPolicyConfig
from aicd.tools.path_policy import PathPolicy


class FSTool:
    def __init__(self, policy: PathPolicy, cwd: Path | None = None) -> None:
        self._policy = policy
        self._cwd = cwd or Path.cwd()
        self._cfg: PathPolicyConfig = policy._cfg  # noqa: SLF001

    def _p(self, path_str: str) -> Path:
        p = self._policy.resolve(path_str, self._cwd)
        self._policy.check_allowed(p)
        return p

    def list_dir(self, path: str) -> dict[str, Any]:
        d = self._p(path)
        if not d.is_dir():
            return {"ok": False, "error": "not a directory"}
        entries = []
        for name in sorted(d.iterdir(), key=lambda x: (not x.is_dir(), x.name.lower())):
            try:
                st = name.stat()
                entries.append(
                    {
                        "name": name.name,
                        "path": str(name),
                        "is_dir": name.is_dir(),
                        "size": st.st_size if name.is_file() else None,
                    }
                )
            except OSError as e:
                entries.append({"name": name.name, "path": str(name), "error": str(e)})
        return {"ok": True, "path": str(d), "entries": entries}

    def read_file(self, path: str, max_bytes: int | None = None) -> dict[str, Any]:
        p = self._p(path)
        if not p.is_file():
            return {"ok": False, "error": "not a file"}
        limit = max_bytes or self._cfg.max_read_bytes
        data = p.read_bytes()
        truncated = len(data) > limit
        chunk = data[:limit]
        text: str | None
        try:
            text = chunk.decode("utf-8")
        except UnicodeDecodeError:
            text = None
        return {
            "ok": True,
            "path": str(p),
            "truncated": truncated,
            "size": len(data),
            "text": text,
            "binary_preview_hex": None
            if text is not None
            else (chunk[:128].hex() + ("..." if len(chunk) > 128 else "")),
        }

    def read_bytes_full(self, path: str, max_bytes: int | None = None) -> dict[str, Any]:
        p = self._p(path)
        if not p.is_file():
            return {"ok": False, "error": "not a file"}
        limit = max_bytes or self._cfg.max_read_bytes * 10
        data = p.read_bytes()
        if len(data) > limit:
            return {"ok": False, "error": f"file too large ({len(data)} > {limit})"}
        return {"ok": True, "path": str(p), "data": data}

    def write_bytes(self, path: str, data: bytes) -> dict[str, Any]:
        p = self._p(path)
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_bytes(data)
        return {"ok": True, "path": str(p), "bytes": len(data)}

    def write_file(self, path: str, content: str, encoding: str = "utf-8") -> dict[str, Any]:
        p = self._p(path)
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text(content, encoding=encoding)
        return {"ok": True, "path": str(p), "bytes": len(content.encode(encoding))}

    def delete_path(self, path: str, dry_run: bool = False) -> dict[str, Any]:
        p = self._p(path)
        if dry_run:
            return {"ok": True, "dry_run": True, "would_delete": str(p)}
        if p.is_dir():
            import shutil

            shutil.rmtree(p)
        elif p.is_file():
            p.unlink()
        else:
            return {"ok": False, "error": "path does not exist"}
        return {"ok": True, "deleted": str(p)}

    def mkdir(self, path: str, parents: bool = True) -> dict[str, Any]:
        p = self._p(path)
        p.mkdir(parents=parents, exist_ok=True)
        return {"ok": True, "path": str(p)}

    def move(self, src: str, dest: str) -> dict[str, Any]:
        s = self._p(src)
        d = self._p(dest)
        self._policy.check_allowed(d)
        d.parent.mkdir(parents=True, exist_ok=True)
        os.replace(s, d)
        return {"ok": True, "from": str(s), "to": str(d)}

    def glob_names(self, directory: str, pattern: str) -> dict[str, Any]:
        d = self._p(directory)
        if not d.is_dir():
            return {"ok": False, "error": "not a directory"}
        matches = []
        for root, _dirs, files in os.walk(d):
            rp = Path(root)
            for f in files:
                if fnmatch.fnmatch(f, pattern):
                    matches.append(str(rp / f))
        return {"ok": True, "count": len(matches), "paths": matches[:500]}

    def search_content_regex(
        self,
        directory: str,
        pattern: str,
        *,
        glob_filter: str = "*",
        max_files: int = 200,
        max_matches_per_file: int = 50,
    ) -> dict[str, Any]:
        d = self._p(directory)
        if not d.is_dir():
            return {"ok": False, "error": "not a directory"}
        try:
            rx = re.compile(pattern)
        except re.error as e:
            return {"ok": False, "error": f"invalid regex: {e}"}
        results: list[dict[str, Any]] = []
        files_seen = 0
        for root, _dirs, files in os.walk(d):
            for fname in files:
                if not fnmatch.fnmatch(fname, glob_filter):
                    continue
                fp = Path(root) / fname
                self._policy.check_allowed(fp)
                if not fp.is_file():
                    continue
                try:
                    if fp.stat().st_size > self._cfg.max_search_file_bytes:
                        continue
                except OSError:
                    continue
                files_seen += 1
                if files_seen > max_files:
                    break
                try:
                    raw = fp.read_bytes()
                    if b"\x00" in raw[:8192]:
                        continue
                    text = raw.decode("utf-8", errors="replace")
                except OSError:
                    continue
                file_matches: list[dict[str, Any]] = []
                for m in rx.finditer(text):
                    file_matches.append(
                        {
                            "start": m.start(),
                            "end": m.end(),
                            "match": m.group(0)[:200],
                        }
                    )
                    if len(file_matches) >= max_matches_per_file:
                        break
                if file_matches:
                    results.append({"path": str(fp), "matches": file_matches})
            if files_seen > max_files:
                break
        return {"ok": True, "results": results, "files_scanned": files_seen}



from __future__ import annotations

from aicd.kit import math_stats


def test_describe() -> None:
    r = math_stats.describe([1.0, 2.0, 3.0])
    assert r["ok"] is True
    assert r["n"] == 3
    assert r["mean"] == 2.0


def test_histogram() -> None:
    r = math_stats.histogram([0.0, 0.1, 9.9, 10.0], bins=2)
    assert r["ok"] is True
    assert r["n_bins"] == 2
    assert sum(r["counts"]) == 4


def test_correlation() -> None:
    r = math_stats.correlation_pearson([1, 2, 3], [2, 4, 6])
    assert r["ok"] is True
    assert abs(float(r["pearson_r"]) - 1.0) < 1e-9


def test_run_operation_describe() -> None:
    r = math_stats.run_operation("describe", {"values": [10, 20]})
    assert r["ok"] is True
    assert r["n"] == 2



from __future__ import annotations

from aicd.kit import text_regex


def test_findall() -> None:
    r = text_regex.regex_findall("a1b2", r"\d")
    assert r["ok"] is True
    assert r["matches"] == ["1", "2"]


def test_sub() -> None:
    r = text_regex.regex_sub("a-b", r"-", "_")
    assert r["ok"] is True
    assert r["text"] == "a_b"


def test_normalize() -> None:
    r = text_regex.normalize_whitespace("  a \n b  ")
    assert r["ok"] is True
    assert r["text"] == "a b"


def test_split_lines() -> None:
    r = text_regex.split_lines("a\nb")
    assert r["ok"] is True
    assert r["lines"] == ["a", "b"]



"""Per-async-task context for nested `task_ai` workers (parent links, scratch paths, messaging)."""

from __future__ import annotations

from contextvars import ContextVar

current_task_id: ContextVar[str | None] = ContextVar("current_task_id", default=None)



from __future__ import annotations

import inspect
import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Mapping

from aicd.agent.handlers.registry import tool_handlers
from aicd.config import AppConfig
from aicd.tasks.manager import TaskManager
from aicd.tools.browser_cdp import BrowserCDP
from aicd.tools.docs_convert import DocumentConverter
from aicd.tools.fs_tool import FSTool
from aicd.tools.http_tool import HttpTool
from aicd.tools.path_policy import PathPolicy
from aicd.tools.process_tool import ProcessTool
from aicd.tools.script_tool import ScriptTool
from aicd.tools.text_tool import TextTool

if TYPE_CHECKING:
    from aicd.session_log import SessionLogger

EnvCapabilitiesCallback = Callable[[dict[str, Any], Path], None]
AiWorkspaceSetter = Callable[[str | None], tuple[bool, str]]


class ToolRouter:
    def __init__(
        self,
        cfg: AppConfig,
        *,
        layout: dict[str, Path],
        task_manager: TaskManager,
        cwd: Path | None = None,
        path_policy: PathPolicy | None = None,
        session_log: SessionLogger | None = None,
        chat_session_id: str | None = None,
        chat_session_switch: Callable[[str], None] | None = None,
    ) -> None:
        self._cfg = cfg
        self._layout = layout
        self._tasks = task_manager
        self._session_log = session_log
        self._chat_session_id = chat_session_id
        self._chat_session_switch = chat_session_switch
        self._cwd = cwd or Path.cwd()
        self._policy = path_policy if path_policy is not None else PathPolicy(cfg.paths)
        self._policy.set_workspace_root(self._cwd)
        self._fs = FSTool(self._policy, self._cwd)
        self._text = TextTool(self._fs)
        self._process = ProcessTool(self._policy, self._cwd)
        self._http = HttpTool(layout["cache"])
        self._scripts = ScriptTool(layout["tmp_scripts"], self._process)
        self._docs = DocumentConverter(lambda p: self._policy.check_allowed(p))
        self._browser = BrowserCDP()
        self._ai_output = (self._cwd / "aicd").resolve()
        self._ai_tmp = (self._ai_output / "tmp").resolve()
        self._handlers = tool_handlers()
        self._on_env_capabilities_updated: EnvCapabilitiesCallback | None = None
        self._ai_workspace_setter: AiWorkspaceSetter | None = None
        self._vision_llm: Any = None

    def set_env_capabilities_callback(
        self, fn: EnvCapabilitiesCallback | None
    ) -> None:
        """主 Agent 在 runtime_env_refresh 后更新系统提示中的能力块。"""
        self._on_env_capabilities_updated = fn

    def notify_env_capabilities_updated(
        self, data: Mapping[str, Any], yaml_path: Path
    ) -> None:
        fn = self._on_env_capabilities_updated
        if fn:
            fn(dict(data), yaml_path)

    def set_chat_session_switch(self, fn: Callable[[str], None] | None) -> None:
        """主 Agent 切换会话 id 时同步清空内存上下文（由 AgentLoop 注册）。"""
        self._chat_session_switch = fn

    def set_chat_session_id(self, sid: str) -> None:
        self._chat_session_id = sid

    def set_ai_workspace_setter(self, fn: AiWorkspaceSetter | None) -> None:
        """由 RuntimeContext 注册，使主会话中 ai_workspace_set 工具可切换工作目录。"""
        self._ai_workspace_setter = fn

    def set_vision_llm(self, client: Any | None) -> None:
        """注册带 ``async chat_with_images(prompt, paths) -> dict`` 的 LLM 客户端（多模态）。"""
        self._vision_llm = client

    @property
    def cwd(self) -> Path:
        return self._cwd

    @property
    def path_policy(self) -> PathPolicy:
        return self._policy

    def get_ai_paths(self) -> tuple[Path, Path]:
        return self._ai_output, self._ai_tmp

    def set_ai_paths(self, output_dir: Path, tmp_dir: Path) -> None:
        """AI 产出根目录与其下 tmp（中间数据）；与 fs 的 cwd 独立。"""
        self._ai_output = output_dir.resolve()
        self._ai_tmp = tmp_dir.resolve()

    def set_cwd(self, cwd: Path) -> None:
        """切换主 Agent 工具使用的当前工作目录（相对路径、默认 shell cwd）。"""
        self._cwd = cwd.resolve()
        self._policy.set_workspace_root(self._cwd)
        self._fs = FSTool(self._policy, self._cwd)
        self._text = TextTool(self._fs)
        self._process = ProcessTool(self._policy, self._cwd)
        self._scripts = ScriptTool(self._layout["tmp_scripts"], self._process)

    def _resolve_path(self, path_str: str) -> Path:
        p = Path(path_str).expanduser()
        if not p.is_absolute():
            p = (self._cwd / p).resolve()
        else:
            p = p.resolve()
        self._policy.check_allowed(p)
        return p

    async def run(self, name: str, arguments: str | dict[str, Any]) -> dict[str, Any]:
        slog = self._session_log
        try:
            args = (
                json.loads(arguments) if isinstance(arguments, str) else dict(arguments)
            )
        except json.JSONDecodeError as e:
            if slog:
                raw = arguments if isinstance(arguments, str) else json.dumps(arguments)
                slog.write_event(
                    "tool",
                    "ToolRouter",
                    name,
                    str(raw)[:800],
                    f"bad_json err={e}",
                )
            return {"ok": False, "error": f"bad json: {e}"}
        try:
            result = await self._dispatch(name, args)
            if slog:
                from aicd.session_log import summarize_tool_call, summarize_tool_result

                try:
                    sin = summarize_tool_call(name, args)
                    if isinstance(result, dict):
                        sout = summarize_tool_result(name, result)
                    else:
                        raw = json.dumps(result, ensure_ascii=False, default=str)
                        sout = raw if len(raw) <= 4000 else raw[:3997] + "..."
                except Exception as ex:  # noqa: BLE001
                    sin = repr(args)[:2000]
                    sout = f"summarize_error={ex!r} raw={str(result)[:1500]}"
                slog.write_event("tool", "ToolRouter", name, sin, sout)
            return result
        except PermissionError as e:
            if slog:
                slog.write_exception("tool", "ToolRouter", name, e)
            return {"ok": False, "error": str(e)}
        except Exception as e:  # noqa: BLE001
            if slog:
                slog.write_exception("tool", "ToolRouter", name, e)
            return {"ok": False, "error": repr(e)}

    async def _dispatch(self, name: str, args: dict[str, Any]) -> dict[str, Any]:
        fn = self._handlers.get(name)
        if fn is None:
            return {"ok": False, "error": f"unknown tool: {name}"}
        result = fn(self, args)
        if inspect.isawaitable(result):
            return await result
        return result



"""将 ``.docx`` 转为 ``.pdf``（LibreOffice 优先，可选 docx2pdf + Word）。"""

from __future__ import annotations

import os
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Any, Callable

_CONVERT_TIMEOUT_SEC = 180


def install_suggestions_docx_to_pdf() -> list[dict[str, Any]]:
    """
    缺失转换引擎时返回给 Agent 的结构化安装建议。

    **约定**：由主对话先 **询问用户是否同意**，再使用 **`run_command`** 执行
    ``shell_command``；勿在未确认时静默安装。
    """
    s: list[dict[str, Any]] = []
    if sys.platform == "win32":
        s.append(
            {
                "id": "libreoffice_winget",
                "title": "安装 LibreOffice（推荐，DOCX→PDF 通用）",
                "shell_command": "winget install --id TheDocumentFoundation.LibreOffice -e --accept-package-agreements",
                "note": "安装完成后**新开终端**或重启 Aicd，再调用 docx_export_pdf。也可从 https://www.libreoffice.org/download/download/ 手动安装。",
            }
        )
        s.append(
            {
                "id": "env_soffice",
                "title": "已安装 LibreOffice 但未在 PATH 中",
                "shell_command": "",
                "note": "在系统环境变量中把 LibreOffice\\program 加入 PATH，或设置环境变量 AICD_SOFFICE 指向 soffice.exe 的完整路径。",
            }
        )
        s.append(
            {
                "id": "pip_docx2pdf_win",
                "title": "可选：pip 安装 docx2pdf（需本机已装 Microsoft Word）",
                "shell_command": 'pip install "aicd[pdf]"',
                "note": "仅在有 Word 时有效；无 Word 请优先 LibreOffice 方案。",
            }
        )
    elif sys.platform == "darwin":
        s.append(
            {
                "id": "libreoffice_brew",
                "title": "安装 LibreOffice（推荐）",
                "shell_command": "brew install --cask libreoffice",
                "note": "无 Homebrew 可从 https://www.libreoffice.org/download/download/ 下载安装。",
            }
        )
        s.append(
            {
                "id": "pip_docx2pdf_mac",
                "title": "可选：pip 安装 docx2pdf（需 Microsoft Word for Mac）",
                "shell_command": 'pip install "aicd[pdf]"',
                "note": "依赖 Word；否则请用 LibreOffice。",
            }
        )
    else:
        s.append(
            {
                "id": "libreoffice_apt",
                "title": "安装 LibreOffice（推荐，Debian/Ubuntu 示例）",
                "shell_command": "sudo apt-get update && sudo apt-get install -y libreoffice",
                "note": "其他发行版请用对应包管理器安装 libreoffice / soffice。",
            }
        )
    s.append(
        {
            "id": "use_aicd_tool",
            "title": "勿手写 reportlab/inline_images 等脚本转 PDF",
            "shell_command": "",
            "note": "请使用内置工具 **docx_export_pdf**；python-docx 的 Document 无 inline_images 属性，易报错。",
        }
    )
    return s


def _attach_suggestions(d: dict[str, Any]) -> dict[str, Any]:
    """为失败结果附加 install_suggestions 与 Agent 指引（原地修改并返回）。"""
    if d.get("ok") is not False:
        return d
    d.setdefault("install_suggestions", install_suggestions_docx_to_pdf())
    d.setdefault(
        "next_step_for_agent",
        "先向用户说明缺少 PDF 转换组件，列出 install_suggestions；**询问用户是否同意安装**后再用 run_command 执行其中的 shell_command（跳过 command 为空的项）；不要未经确认就执行安装。",
    )
    return d


def find_soffice_executable() -> Path | None:
    """
    查找 LibreOffice ``soffice`` / ``libreoffice``。

    环境变量 **AICD_SOFFICE** 或 **SOFFICE_PATH** 可指向可执行文件（含 ``soffice.exe``）。
    """
    for key in ("AICD_SOFFICE", "SOFFICE_PATH"):
        env = (os.environ.get(key) or "").strip()
        if env:
            p = Path(env).expanduser()
            if p.is_file():
                return p.resolve()
    for name in ("soffice", "libreoffice"):
        w = shutil.which(name)
        if w:
            return Path(w).resolve()
    if sys.platform == "win32":
        for pf_key in ("ProgramFiles", "ProgramFiles(x86)"):
            pf = os.environ.get(pf_key) or ""
            if not pf:
                continue
            root = Path(pf)
            if not root.is_dir():
                continue
            for lo in sorted(root.glob("LibreOffice*/program/soffice.exe")):
                if lo.is_file():
                    return lo.resolve()
    if sys.platform == "darwin":
        mac = Path("/Applications/LibreOffice.app/Contents/MacOS/soffice")
        if mac.is_file():
            return mac.resolve()
    return None


def _export_via_libreoffice(
    docx_path: Path,
    output_pdf: Path,
) -> dict[str, Any]:
    exe = find_soffice_executable()
    if exe is None:
        return {
            "ok": False,
            "error": "LibreOffice (soffice) not found",
            "hint": "Install LibreOffice and ensure soffice is on PATH, or set AICD_SOFFICE to the executable.",
        }
    out_dir = output_pdf.parent
    out_dir.mkdir(parents=True, exist_ok=True)
    expected = out_dir / f"{docx_path.stem}.pdf"
    # 若目标与 LO 默认输出名不同，先写到 expected 再移动
    cmd = [
        str(exe),
        "--headless",
        "--norestore",
        "--nofirststartwizard",
        "--convert-to",
        "pdf",
        "--outdir",
        str(out_dir),
        str(docx_path),
    ]
    try:
        cp = subprocess.run(
            cmd,
            capture_output=True,
            text=True,
            timeout=_CONVERT_TIMEOUT_SEC,
            encoding="utf-8",
            errors="replace",
        )
    except subprocess.TimeoutExpired:
        return {
            "ok": False,
            "error": f"LibreOffice timed out after {_CONVERT_TIMEOUT_SEC}s",
            "via": "libreoffice",
        }
    except OSError as e:
        return {"ok": False, "error": f"LibreOffice spawn failed: {e}", "via": "libreoffice"}

    if not expected.is_file() or expected.stat().st_size == 0:
        tail = (cp.stderr or cp.stdout or "")[:2000]
        return {
            "ok": False,
            "error": f"LibreOffice did not produce {expected.name} (exit {cp.returncode})",
            "stderr_tail": tail,
            "via": "libreoffice",
        }

    if expected.resolve() != output_pdf.resolve():
        try:
            if output_pdf.is_file():
                output_pdf.unlink()
            shutil.move(str(expected), str(output_pdf))
        except OSError as e:
            return {
                "ok": False,
                "error": f"could not move PDF to destination: {e}",
                "via": "libreoffice",
            }

    return {
        "ok": True,
        "pdf_path": str(output_pdf),
        "via": "libreoffice",
        "soffice": str(exe),
    }


def _export_via_docx2pdf(docx_path: Path, output_pdf: Path) -> dict[str, Any]:
    if sys.platform not in ("win32", "darwin"):
        return {
            "ok": False,
            "error": "docx2pdf is only supported on Windows and macOS (requires Microsoft Word)",
            "via": "docx2pdf",
        }
    try:
        from docx2pdf import convert
    except ImportError:
        return {
            "ok": False,
            "error": "docx2pdf not installed",
            "hint": 'pip install "aicd[pdf]"  （可选；需本机已安装 Microsoft Word）',
            "via": "docx2pdf",
        }
    output_pdf.parent.mkdir(parents=True, exist_ok=True)
    try:
        convert(str(docx_path), str(output_pdf))
    except Exception as e:  # noqa: BLE001
        return {
            "ok": False,
            "error": repr(e),
            "hint": "Ensure Microsoft Word is installed and not showing a blocking dialog.",
            "via": "docx2pdf",
        }
    if not output_pdf.is_file() or output_pdf.stat().st_size == 0:
        return {
            "ok": False,
            "error": "docx2pdf did not write a non-empty PDF",
            "via": "docx2pdf",
        }
    return {"ok": True, "pdf_path": str(output_pdf), "via": "docx2pdf"}


def export_docx_to_pdf(
    docx_path: Path,
    output_pdf: Path | None,
    *,
    policy_check: Callable[[Path], None],
    prefer: str | None = None,
) -> dict[str, Any]:
    """
    ``prefer``: ``libreoffice`` | ``docx2pdf`` | ``None``（自动：先 LO，再在 Win/Mac 上尝试 docx2pdf）。
    """
    src = docx_path.expanduser().resolve()
    policy_check(src)
    if not src.is_file():
        return {"ok": False, "error": f"docx file not found: {src}"}
    if src.suffix.lower() != ".docx":
        return {"ok": False, "error": "source must be a .docx file"}

    dst = (
        output_pdf.expanduser().resolve()
        if output_pdf is not None
        else src.with_suffix(".pdf")
    )
    if dst.suffix.lower() != ".pdf":
        return {"ok": False, "error": "output path must end with .pdf"}
    policy_check(dst)
    policy_check(dst.parent)

    pref = (prefer or "").strip().lower()
    attempts: list[tuple[str, Callable[[], dict[str, Any]]]] = []

    if pref == "docx2pdf":
        attempts.append(("docx2pdf", lambda: _export_via_docx2pdf(src, dst)))
        attempts.append(("libreoffice", lambda: _export_via_libreoffice(src, dst)))
    else:
        attempts.append(("libreoffice", lambda: _export_via_libreoffice(src, dst)))
        if pref != "libreoffice":
            attempts.append(("docx2pdf", lambda: _export_via_docx2pdf(src, dst)))

    errors: list[str] = []
    hints: list[str] = []
    for name, fn in attempts:
        r = fn()
        if r.get("ok"):
            r["source_docx"] = str(src)
            return r
        msg = str(r.get("error") or "failed")
        errors.append(f"{name}: {msg}")
        h = r.get("hint")
        if isinstance(h, str) and h.strip():
            hints.append(h.strip())

    out: dict[str, Any] = {
        "ok": False,
        "error": " | ".join(errors),
        "source_docx": str(src),
    }
    if hints:
        out["hint"] = " ".join(dict.fromkeys(hints))
    else:
        out["hint"] = (
            "Install LibreOffice (recommended) or on Windows/macOS install Microsoft Word and pip install docx2pdf."
        )
    return _attach_suggestions(out)



"""diagrams.net / Draw.io（.drawio）XML 生成与轻量解析。"""

from __future__ import annotations

import html
import re
import uuid
import xml.etree.ElementTree as ET
from datetime import datetime, timezone
from typing import Any

MAX_NODES = 800
MAX_EDGES = 2000
MAX_XML_CHARS = 4_000_000


def _utc_ts() -> str:
    return datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S.000Z")


def _esc_attr(s: str) -> str:
    return (
        html.escape(s, quote=True)
        .replace("\n", "&#xa;")
        .replace("\r", "")
    )


def build_drawio_xml(
    nodes: list[dict[str, Any]],
    edges: list[dict[str, Any]],
    *,
    page_name: str = "Page-1",
    diagram_id: str | None = None,
    cols: int = 4,
    cell_width: int = 140,
    cell_height: int = 72,
    grid_x: int = 48,
    grid_y: int = 48,
    x_gap: int = 200,
    y_gap: int = 120,
) -> dict[str, Any]:
    """由简单 nodes/edges 生成未压缩的 .drawio（mxfile）XML 文本。"""
    if not isinstance(nodes, list) or not isinstance(edges, list):
        return {"ok": False, "error": "nodes and edges must be arrays"}
    if len(nodes) > MAX_NODES:
        return {"ok": False, "error": f"at most {MAX_NODES} nodes"}
    if len(edges) > MAX_EDGES:
        return {"ok": False, "error": f"at most {MAX_EDGES} edges"}
    cols = max(1, min(12, int(cols)))

    id_map: dict[str, str] = {}
    next_mx = 2

    norm_nodes: list[tuple[str, str, dict[str, Any]]] = []
    for i, raw in enumerate(nodes):
        if not isinstance(raw, dict):
            return {"ok": False, "error": f"nodes[{i}] must be object"}
        uid = str(raw.get("id") if raw.get("id") is not None else f"n{i}")
        if uid in id_map:
            return {"ok": False, "error": f"duplicate node id: {uid!r}"}
        mid = str(next_mx)
        next_mx += 1
        id_map[uid] = mid
        lbl = str(raw.get("label") if raw.get("label") is not None else uid)
        norm_nodes.append((uid, mid, {**raw, "label": lbl}))

    for i, e in enumerate(edges):
        if not isinstance(e, dict):
            return {"ok": False, "error": f"edges[{i}] must be object"}
        s = str(e.get("source", e.get("from", "")))
        t = str(e.get("target", e.get("to", "")))
        if s not in id_map or t not in id_map:
            return {
                "ok": False,
                "error": f"edges[{i}] unknown source/target: {s!r} -> {t!r}",
            }

    did = diagram_id or str(uuid.uuid4())
    page_name = (page_name or "Page-1").strip() or "Page-1"

    parts: list[str] = []
    parts.append(
        f'<mxfile host="app.diagrams.net" modified="{_utc_ts()}" '
        f'agent="Aicd" version="22.1.0" etag="{uuid.uuid4()}" type="device">'
    )
    parts.append(f'  <diagram id="{_esc_attr(did)}" name="{_esc_attr(page_name)}">')
    parts.append(
        '    <mxGraphModel dx="1200" dy="800" grid="1" gridSize="10" guides="1" '
        'tooltips="1" connect="1" arrows="1" fold="1" page="1" pageScale="1" '
        'pageWidth="1169" pageHeight="827" math="0" shadow="0">'
    )
    parts.append("      <root>")
    parts.append('        <mxCell id="0"/>')
    parts.append('        <mxCell id="1" parent="0"/>')

    for idx, (_uid, mid, meta) in enumerate(norm_nodes):
        lbl = _esc_attr(str(meta["label"]))
        style = str(meta.get("style") or "rounded=1;whiteSpace=wrap;html=1;")
        if "fillColor" not in style and meta.get("fill_color"):
            style += f"fillColor={meta['fill_color']};"
        x = meta.get("x")
        y = meta.get("y")
        w = int(meta.get("width") or cell_width)
        h = int(meta.get("height") or cell_height)
        if x is None or y is None:
            c = idx % cols
            r = idx // cols
            x = grid_x + c * x_gap
            y = grid_y + r * y_gap
        else:
            x, y = int(x), int(y)
        parts.append(
            f'        <mxCell id="{mid}" value="{lbl}" style="{_esc_attr(style)}" '
            f'vertex="1" parent="1">'
            f'<mxGeometry x="{x}" y="{y}" width="{w}" height="{h}" as="geometry"/>'
            f"</mxCell>"
        )

    edge_style = (
        "edgeStyle=orthogonalEdgeStyle;rounded=0;orthogonalLoop=1;"
        "jettySize=auto;html=1;"
    )
    for e in edges:
        s = str(e.get("source", e.get("from", "")))
        t = str(e.get("target", e.get("to", "")))
        sid = id_map[s]
        tid = id_map[t]
        eid = str(next_mx)
        next_mx += 1
        elbl = e.get("label")
        val = _esc_attr(str(elbl)) if elbl is not None and str(elbl) != "" else ""
        val_attr = f' value="{val}"' if val else ""
        parts.append(
            f'        <mxCell id="{eid}"{val_attr} style="{edge_style}" edge="1" '
            f'parent="1" source="{sid}" target="{tid}">'
            '<mxGeometry relative="1" as="geometry"/>'
            "</mxCell>"
        )

    parts.append("      </root>")
    parts.append("    </mxGraphModel>")
    parts.append("  </diagram>")
    parts.append("</mxfile>")
    xml = "\n".join(parts)
    return {
        "ok": True,
        "xml": xml,
        "node_count": len(norm_nodes),
        "edge_count": len(edges),
        "diagram_id": did,
    }


def _local_tag(tag: str) -> str:
    if "}" in tag:
        return tag.split("}", 1)[1]
    return tag


def summarize_drawio_xml(xml_text: str) -> dict[str, Any]:
    """从 .drawio 文本中提取顶点/边摘要（最佳努力，不校验全量 schema）。"""
    if len(xml_text) > MAX_XML_CHARS:
        return {"ok": False, "error": f"XML exceeds {MAX_XML_CHARS} chars"}
    try:
        root = ET.fromstring(xml_text)
    except ET.ParseError as e:
        return {"ok": False, "error": str(e)}

    nodes: list[dict[str, Any]] = []
    edges: list[dict[str, Any]] = []
    for el in root.iter():
        if _local_tag(el.tag) != "mxCell":
            continue
        cid = el.get("id")
        if el.get("vertex") == "1":
            geom = el.find("mxGeometry")
            x = y = w = h = None
            if geom is not None:
                x = geom.get("x")
                y = geom.get("y")
                w = geom.get("width")
                h = geom.get("height")
            nodes.append(
                {
                    "cell_id": cid,
                    "label": el.get("value") or "",
                    "style": el.get("style") or "",
                    "x": x,
                    "y": y,
                    "width": w,
                    "height": h,
                }
            )
        elif el.get("edge") == "1":
            edges.append(
                {
                    "cell_id": cid,
                    "source": el.get("source"),
                    "target": el.get("target"),
                    "label": el.get("value") or "",
                }
            )

    return {
        "ok": True,
        "vertices": nodes,
        "edges": edges,
        "vertex_count": len(nodes),
        "edge_count": len(edges),
    }


def strip_bom_and_xml_decl(s: str) -> str:
    s = s.lstrip("\ufeff")
    s = re.sub(r"^\s*<\?xml[^>]*\?>\s*", "", s, count=1, flags=re.I)
    return s.strip()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from aicd.tools.script_backup import policy_allow_under_ai_output_or_workspace

from aicd.agent.tool_args_coerce import (
    coerce_tool_json_array,
    coerce_tool_json_object,
)
from aicd.tools.process_tool import (
    shell_argv_from_user_input,
    warn_shell_preflight,
)

if TYPE_CHECKING:
    from aicd.agent.tool_router import ToolRouter


def _text_regex_replace(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    return router._text.regex_replace(
        args["path"],
        args["pattern"],
        args["replacement"],
        count=int(args.get("count") or 0),
    )


def _text_patch_bytes_hex(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    return router._text.patch_bytes(
        args["path"], int(args["offset"]), args["hex_data"]
    )


async def _run_command(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    timeout = float(args.get("timeout_sec", 600))
    raw_argv = args.get("argv")
    shell = args.get("shell")
    if raw_argv is not None:
        co = coerce_tool_json_array(raw_argv, "argv", allow_empty=True)
        if isinstance(co, dict):
            return co
        raw_argv = co
    cmd = args.get("command")
    if raw_argv is None and cmd is not None and str(cmd).strip():
        if (w := warn_shell_preflight(str(cmd))):
            return {"ok": False, "error": w}
    try:
        argv = shell_argv_from_user_input(
            command=args.get("command"),
            argv=[str(x) for x in raw_argv] if raw_argv else None,
            shell=str(shell).strip() if shell is not None else None,
        )
    except ValueError as e:
        return {"ok": False, "error": str(e)}

    async def _log(line: str) -> None:
        await router._tasks.bus.publish(
            {"topic": "agent", "event": "log", "line": line}
        )

    return await router._process.run_command(
        argv,
        cwd=args.get("cwd"),
        timeout_sec=timeout,
        on_line=_log,
    )


async def _http_request(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    hdrs = args.get("headers")
    if hdrs is not None:
        h_obj, h_err = coerce_tool_json_object(hdrs, "headers")
        if h_err:
            return {"ok": False, "error": h_err}
        hdrs = h_obj
    jb = args.get("json_body")
    if jb is not None:
        j_obj, j_err = coerce_tool_json_object(jb, "json_body")
        if j_err:
            return {"ok": False, "error": j_err}
        jb = j_obj
    return await router._http.request(
        args["method"],
        args["url"],
        headers=hdrs,
        json_body=jb,
        data=args.get("data"),
        timeout_sec=float(args.get("timeout_sec", 60)),
    )


async def _http_download(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    return await router._http.download(
        args["url"], filename=args.get("filename")
    )


async def _script_run(router: ToolRouter, args: dict[str, Any]) -> dict[str, Any]:
    kind = str(args["kind"]).lower()
    src = args["source"]
    timeout = float(args.get("timeout_sec", 300))
    rd = args.get("related_doc")
    related_doc = str(rd).strip() if isinstance(rd, str) and rd.strip() else None
    bak_root = (router._ai_output / "bak").resolve()
    bak_policy = policy_allow_under_ai_output_or_workspace(
        router._ai_output, router._policy.check_allowed
    )

    async def _log(line: str) -> None:
        await router._tasks.bus.publish(
            {"topic": "agent", "event": "log", "line": line}
        )

    common = dict(
        on_line=_log,
        timeout_sec=timeout,
        related_doc=related_doc,
        backup_dir=bak_root,
        policy_check=bak_policy,
    )
    if kind == "py":
        return await router._scripts.run_python(src, **common)
    if kind == "ps1":
        return await router._scripts.run_powershell(src, **common)
    if kind == "cmd":
        return await router._scripts.run_cmd(src, **common)
    if kind == "bash":
        return await router._scripts.run_bash(src, **common)
    return {"ok": False, "error": f"unknown script kind: {kind}"}


HANDLERS: dict[str, Any] = {
    "text_regex_replace": _text_regex_replace,
    "text_patch_bytes_hex": _text_patch_bytes_hex,
    "run_command": _run_command,
    "http_request": _http_request,
    "http_download": _http_download,
    "script_run": _script_run,
}



from __future__ import annotations

import os
from pathlib import Path

from aicd.config import PathPolicyConfig


class PathPolicy:
    def __init__(self, cfg: PathPolicyConfig) -> None:
        self._cfg = cfg
        self._roots = [_expand_root(r) for r in cfg.allowed_roots]
        # 当前 AI 工作区（随 /work 变化）；full_disk=false 时始终允许该目录下读写
        self._workspace_root: Path | None = None

    def set_workspace_root(self, root: Path | None) -> None:
        """与 ToolRouter 主工作目录同步；解决 allowed_roots 里 "." 只在启动时展开的问题。"""
        self._workspace_root = root.resolve() if root is not None else None

    def resolve(self, path_str: str, cwd: Path | None = None) -> Path:
        base = cwd or Path.cwd()
        p = Path(path_str).expanduser()
        if not p.is_absolute():
            p = (base / p).resolve()
        else:
            p = p.resolve()
        return p

    def check_allowed(self, path: Path) -> None:
        if self._cfg.full_disk:
            return
        path = path.resolve()
        wr = self._workspace_root
        if wr is not None:
            try:
                path.relative_to(wr)
                return
            except ValueError:
                pass
        for root in self._roots:
            try:
                path.relative_to(root)
                return
            except ValueError:
                continue
        raise PermissionError(f"Path not under allowed_roots: {path}")


def _expand_root(spec: str) -> Path:
    s = os.path.expandvars(spec)
    return Path(s).expanduser().resolve()



from __future__ import annotations

import asyncio
from pathlib import Path


def test_viz_data_chart_and_data_stats(tmp_path: Path) -> None:
    from aicd.runtime import build_runtime

    async def _go() -> None:
        ctx = await build_runtime(home=tmp_path / "home")
        ok, _ = ctx.set_ai_workspace(str(tmp_path))
        assert ok is True
        w = await ctx.router.run("ai_workspace_set", {})
        assert w.get("ok") is True
        assert Path(w["workspace"]) == tmp_path.resolve()
        sub = tmp_path / "nested"
        sub.mkdir()
        w2 = await ctx.router.run("ai_workspace_set", {"path": "nested"})
        assert w2.get("ok") is True
        assert Path(w2["workspace"]) == sub.resolve()
        html = tmp_path / "c.html"
        r = await ctx.router.run(
            "viz_data_chart",
            {
                "html_path": str(html),
                "labels": ["a", "b"],
                "values": [1, 3],
                "chart_type": "bar",
            },
        )
        assert r.get("ok") is True
        assert html.is_file()
        assert (html.parent / "res" / "chart.umd.min.js").is_file()

        s = await ctx.router.run(
            "data_stats",
            {"operation": "describe", "values": [1, 2, 3]},
        )
        assert s.get("ok") is True
        assert s.get("mean") == 2.0
        await ctx.conn.close()

    asyncio.run(_go())
